import os
from typing import Union
import numpy as np
from datetime import datetime

from dliswriter import (
    DLISFile,
    LogicalFile,
)
from dliswriter.logical_record import eflr_types
from dliswriter.utils.enums import (
    Unit,
    Property,
    ZoneDomain,
    EquipmentType,
    EquipmentLocation,
    CalibrationMeasurementPhase,
    ProcessStatus,
)

from tests.dlis_files_for_testing.common import make_file_header, make_sul


def _add_origin(lf: LogicalFile) -> eflr_types.OriginItem:
    origin = lf.add_origin(
        "DEFAULT ORIGIN",
        creation_time="2050/03/02 15:30:00",
        file_set_name="Test file set name",
        file_set_number=42,
        origin_reference=42,
        file_number=8,
        run_number=13,
        well_id=5,
        well_name="Test well name",
        field_name="Test field name",
        company="Test company",
    )

    return origin


def _add_frame(
    lf: LogicalFile, *channels: eflr_types.ChannelItem
) -> eflr_types.FrameItem:
    fr = lf.add_frame(
        "MAIN FRAME",
        channels=channels,
        index_type="TIME",
        spacing=0.5,
        encrypted=1,
        description="Frame description",
    )

    fr.spacing.units = Unit.SECOND
    return fr


def _add_channels(
    lf: LogicalFile, ax1: eflr_types.AxisItem, ln: eflr_types.LongNameItem
) -> tuple[eflr_types.ChannelItem, ...]:
    ch = lf.add_channel(
        name="Some Channel",
        dataset_name="image1",
        long_name="Some not so very long channel name",
        properties=[
            Property.AVERAGED,
            Property.LOCALLY_DEFINED,
            Property.SPEED_CORRECTED,
        ],
        cast_dtype=np.float32,
        units=Unit.ACRE,
        dimension=12,
        axis=ax1,
        element_limit=12,
        minimum_value=0,
        maximum_value=127.6,
    )

    ch1 = lf.add_channel(name="Channel 1", dimension=[10, 10], units=Unit.INCH)
    ch2 = lf.add_channel("Channel 2", long_name=ln)
    ch3 = lf.add_channel("Channel 13", dataset_name="amplitude", element_limit=128)
    ch_time = lf.add_channel(
        "posix time", dataset_name="contents/time", units=Unit.SECOND
    )
    ch_rpm = lf.add_channel("surface rpm", dataset_name="contents/rpm", long_name=ln)
    ch_amplitude = lf.add_channel(
        "amplitude", dataset_name="contents/image0", dimension=128
    )
    ch_radius = lf.add_channel(
        "radius", dataset_name="contents/image1", dimension=128, units="in"
    )
    ch_radius_pooh = lf.add_channel(
        "radius_pooh", dataset_name="contents/image2", units=Unit.METER
    )
    ch_x = lf.add_channel(
        "channel_x",
        long_name="Channel not added to the frame",
        dataset_name="image2",
        units="s",
    )

    return (
        ch,
        ch1,
        ch2,
        ch3,
        ch_time,
        ch_rpm,
        ch_amplitude,
        ch_radius,
        ch_radius_pooh,
        ch_x,
    )


def _add_axes(lf: LogicalFile) -> tuple[eflr_types.AxisItem, ...]:
    ax1 = lf.add_axis(
        name="Axis-1", axis_id="First axis", coordinates=list(range(12)), spacing=1
    )
    ax1.spacing.units = Unit.METER

    ax2 = lf.add_axis(
        "Axis-X", axis_id="Axis not added to computation", coordinates=[8], spacing=2
    )
    ax2.spacing.units = "m"

    ax3 = lf.add_axis(name="Axis-3", axis_id="Whatever axis", coordinates=[1, 5.5, 8])

    ax4 = lf.add_axis(name="Axis-4", axis_id="Whatever axis nr 2", coordinates=[-1, 0])

    return ax1, ax2, ax3, ax4


def _add_zones(lf: LogicalFile) -> tuple[eflr_types.ZoneItem, ...]:
    z1 = lf.add_zone(
        name="Zone-1",
        description="BOREHOLE-DEPTH-ZONE",
        domain="BOREHOLE-DEPTH",
        maximum=1300,
        minimum=100,
    )
    z1.maximum.units = "m"
    z1.minimum.units = "m"

    z2 = lf.add_zone(
        "Zone-2",
        description="VERTICAL-DEPTH-ZONE",
        domain=ZoneDomain.VERTICAL_DEPTH,
        maximum=2300.45,
        minimum=200.0,
    )
    z2.maximum.units = Unit.METER
    z2.minimum.units = "m"

    z3 = lf.add_zone(
        "Zone-3",
        description="ZONE-TIME",
        domain=ZoneDomain.TIME,
        maximum="2050/07/13 11:30:00",
        minimum="2050/07/12 9:00:00",
    )

    z4 = lf.add_zone(
        "Zone-4", description="ZONE-TIME-2", domain="TIME", maximum=90, minimum=10
    )
    z4.maximum.units = "min"
    z4.minimum.units = Unit.MINUTE

    zx = lf.add_zone(
        name="Zone-X",
        description="Zone not added to any parameter",
        domain="TIME",
        maximum=10,
        minimum=1,
    )
    zx.maximum.units = Unit.SECOND
    zx.minimum.units = Unit.SECOND

    return z1, z2, z3, z4, zx


def _add_parameters(
    lf: LogicalFile, zones: tuple[eflr_types.ZoneItem, ...], ln: eflr_types.LongNameItem
) -> tuple[eflr_types.ParameterItem, ...]:
    p1 = lf.add_parameter(
        name="Param-1",
        long_name="LATLONG-GPS",
        zones=[zones[0], zones[2]],
        values=["40deg 23' 42.8676'' N", "40deg 23' 42.8676'' N"],
    )

    p2 = lf.add_parameter(
        name="Param-2",
        long_name="LATLONG",
        zones=[zones[1], zones[3]],
        dimension=[3],
        values=[[40.395241, 27.792471, 21.23131213], [21, 23, 24]],
    )

    p3 = lf.add_parameter(name="Param-3", long_name=ln, values=[12.5])
    p3.values.units = Unit.METER

    return p1, p2, p3


def _add_equipment(lf: LogicalFile) -> tuple[eflr_types.EquipmentItem, ...]:
    eq1 = lf.add_equipment(
        name="EQ1",
        trademark_name="EQ-TRADEMARKNAME",
        status=1,
        eq_type=EquipmentType.TOOL,
        serial_number="9101-21391",
        location=EquipmentLocation.WELL,
        height=140,
        length=230.78,
        minimum_diameter=2.3,
        maximum_diameter=3.2,
        volume=100,
        weight=1.2,
        hole_size=323.2,
        pressure=18000,
        temperature=24,
        vertical_depth=587,
        radial_drift=23.22,
        angular_drift=32.5,
    )

    eq1.height.units = Unit.INCH
    eq1.length.units = Unit.CENTIMETER
    eq1.minimum_diameter.units = Unit.METER
    eq1.maximum_diameter.units = Unit.METER
    eq1.weight.units = Unit.METRIC_TON
    eq1.hole_size.units = "m"
    eq1.pressure.units = "psi"
    eq1.temperature.units = Unit.DEGREE_CELSIUS
    eq1.vertical_depth.units = "m"
    eq1.radial_drift.units = "m"
    eq1.angular_drift.units = Unit.METER

    eq2 = lf.add_equipment(
        name="EQ2",
        trademark_name="EQ-TRADEMARKNAME",
        status=0,
        eq_type=EquipmentType.TOOL,
        serial_number="5559101-21391",
    )

    eq3 = lf.add_equipment(
        name="EqX",
        trademark_name="EQ-TRADEMARKNAME",
        status=1,
        eq_type="Tool",
        serial_number="12311",
    )

    return eq1, eq2, eq3


def _add_tools(
    lf: LogicalFile,
    equipment: tuple[eflr_types.EquipmentItem, ...],
    parameters: tuple[eflr_types.ParameterItem, ...],
    channels: tuple[eflr_types.ChannelItem, ...],
) -> tuple[eflr_types.ToolItem, ...]:
    t1 = lf.add_tool(
        name="TOOL-1",
        description="SOME TOOL",
        trademark_name="SMTL",
        generic_name="TOOL GEN NAME",
        parts=[equipment[0], equipment[1]],
        status=1,
        channels=[channels[4], channels[6]],
        parameters=[parameters[0], parameters[2]],
    )

    t2 = lf.add_tool(
        name="Tool-X",
        description="desc",
        trademark_name="SMTL",
        generic_name="TOOL GEN NAME",
        parts=[equipment[1]],
        status=0,
        channels=[channels[8]],
        parameters=[parameters[1]],
    )

    return t1, t2


def _add_processes(
    lf: LogicalFile,
    parameters: tuple[eflr_types.ParameterItem, ...],
    channels: tuple[eflr_types.ChannelItem, ...],
    computations: tuple[eflr_types.ComputationItem, ...],
) -> tuple[eflr_types.ProcessItem, ...]:
    p1 = lf.add_process(
        name="Process 1",
        description="MERGED",
        trademark_name="PROCESS 1",
        version="0.0.1",
        properties=[Property.AVERAGED],
        status="COMPLETE",
        input_channels=[channels[7]],
        output_channels=[channels[6], channels[2]],
        input_computations=[computations[0]],
        output_computations=[computations[1]],
        parameters=parameters,
        comments=["SOME COMMENT HERE"],
    )

    p2 = lf.add_process(
        name="Prc2",
        description="MERGED2",
        trademark_name="PROCESS 2",
        version="0.0.2",
        properties=["AVERAGED"],
        status=ProcessStatus.COMPLETE,
        input_channels=[channels[1]],
        output_channels=[channels[2]],
        input_computations=[computations[1], computations[0]],
        parameters=[parameters[0]],
        comments=["Other comment"],
    )

    return p1, p2


def _add_computation(
    lf: LogicalFile,
    axes: tuple[eflr_types.AxisItem, ...],
    zones: tuple[eflr_types.ZoneItem, ...],
    tools: tuple[eflr_types.ToolItem, ...],
    ln: eflr_types.LongNameItem,
) -> tuple[eflr_types.ComputationItem, ...]:
    c1 = lf.add_computation(
        name="COMPT-1",
        long_name="COMPT1",
        properties=[Property.LOCALLY_DEFINED, "AVERAGED"],
        dimension=[3],
        axis=[axes[2]],
        zones=zones[:2],
        values=[[100, 200, 300], [1, 2, 3]],
        source=tools[0],
    )

    c2 = lf.add_computation(
        name="COMPT2",
        long_name=ln,
        properties=["UNDER-SAMPLED", Property.AVERAGED],
        dimension=[2],
        axis=[axes[3]],
        zones=[zones[0], zones[2]],
        values=[[1.5, 2.5], [4.5, 3.2]],
    )

    cx = lf.add_computation(
        name="COMPT-X",
        long_name="Computation not added to process",
        properties=["OVER-SAMPLED"],
        axis=[axes[0]],
        dimension=[12],
        values=[list(range(12, 24))],
    )

    return c1, c2, cx


def _add_splices(
    lf: LogicalFile,
    channels: tuple[eflr_types.ChannelItem, ...],
    zones: tuple[eflr_types.ZoneItem, ...],
) -> tuple[eflr_types.SpliceItem]:
    s = lf.add_splice(
        name="splc1",
        output_channel=channels[6],
        input_channels=[channels[1], channels[2]],
        zones=[zones[0], zones[1]],
    )

    return (s,)


def _add_calibrations(
    lf: LogicalFile,
    axes: tuple[eflr_types.AxisItem, ...],
    channels: tuple[eflr_types.ChannelItem, ...],
    parameters: tuple[eflr_types.ParameterItem, ...],
) -> tuple[
    eflr_types.CalibrationMeasurementItem,
    eflr_types.CalibrationCoefficientItem,
    eflr_types.CalibrationItem,
]:
    cm = lf.add_calibration_measurement(
        name="CMEASURE-1",
        phase=CalibrationMeasurementPhase.BEFORE,
        axis=axes[3],
        measurement_source=channels[1],
        measurement_type="Plus",
        measurement=[[12.2323, 12.2131]],
        sample_count=12,
        maximum_deviation=[[2.2324, 3.121]],
        standard_deviation=[[1.123, 1.1231]],
        begin_time=datetime(year=2050, month=3, day=12, hour=12, minute=30),
        duration=15,
        reference=[[11, 12]],
        standard=[[11.2, 12.2]],
        plus_tolerance=[[2, 2]],
        minus_tolerance=[[1, 2]],
    )
    cm.duration.units = "s"

    cc = lf.add_calibration_coefficient(
        name="COEF-1",
        label="Gain",
        coefficients=[100.2, 201.3],
        references=[89, 298],
        plus_tolerances=[100.2, 222.124],
        minus_tolerances=[87.23, 214],
    )

    c = lf.add_calibration(
        name="CALIB-MAIN",
        calibrated_channels=[channels[1], channels[2]],
        uncalibrated_channels=[channels[6], channels[7], channels[8]],
        coefficients=[cc],
        measurements=[cm],
        parameters=parameters,
        method="Two Point Linear",
    )

    return cm, cc, c


def _add_well_reference_points(
    lf: LogicalFile,
) -> tuple[eflr_types.WellReferencePointItem, ...]:
    w1 = lf.add_well_reference_point(
        name="AQLN WELL-REF",
        permanent_datum="AQLN permanent_datum",
        vertical_zero="AQLN vertical_zero",
        permanent_datum_elevation=1234.51,
        above_permanent_datum=888.51,
        magnetic_declination=999.51,
        coordinate_1_name="Latitude",
        coordinate_1_value=40.395240,
        coordinate_2_name="Longitude",
        coordinate_2_value=27.792470,
    )

    w2 = lf.add_well_reference_point(
        name="WRP-X",
        permanent_datum="pd1",
        vertical_zero="vz20",
        permanent_datum_elevation=32.5,
        above_permanent_datum=100,
        magnetic_declination=112.3,
        coordinate_1_name="X",
        coordinate_1_value=20,
        coordinate_2_name="Y",
        coordinate_2_value=-0.3,
        coordinate_3_name="Z",
        coordinate_3_value=1,
    )

    return w1, w2


def _add_paths(
    lf: LogicalFile,
    frame: eflr_types.FrameItem,
    wrp: eflr_types.WellReferencePointItem,
    channels: tuple[eflr_types.ChannelItem, ...],
) -> tuple[eflr_types.PathItem, ...]:
    path1 = lf.add_path(
        "PATH-1",
        frame_type=frame,
        well_reference_point=wrp,
        value=(channels[0], channels[1], channels[2]),
        borehole_depth=122.12,
        vertical_depth=211.1,
        radial_drift=12,
        angular_drift=1.11,
        time=13,
    )

    path2 = lf.add_path(
        "PATH-2", value=(channels[4],), tool_zero_offset=1231.1, time=11.1
    )

    return path1, path2


def _add_messages(lf: LogicalFile) -> tuple[eflr_types.MessageItem]:
    m = lf.add_message(
        name="MESSAGE-1",
        message_type="Command",
        time=datetime(year=2050, month=3, day=4, hour=11, minute=23, second=11),
        borehole_drift=123.34,
        vertical_depth=234.45,
        radial_drift=345.56,
        angular_drift=456.67,
        text=["Test message 11111"],
    )

    return (m,)


def _add_comments(lf: LogicalFile) -> tuple[eflr_types.CommentItem, ...]:
    c1 = lf.add_comment(name="COMMENT-1", text=["SOME COMMENT HERE"])

    c2 = lf.add_comment(
        name="cmt2", text=["some other comment here", "and another comment"]
    )

    return c1, c2


def _add_no_formats(lf: LogicalFile) -> tuple[eflr_types.NoFormatItem, ...]:
    nf1 = lf.add_no_format(
        name="no_format_1",
        consumer_name="SOME TEXT NOT FORMATTED",
        description="TESTING-NO-FORMAT",
    )

    nf2 = lf.add_no_format(
        name="no_fmt2", consumer_name="xyz", description="TESTING NO FORMAT 2"
    )

    return nf1, nf2


def _add_long_name(lf: LogicalFile) -> eflr_types.LongNameItem:
    ln = lf.add_long_name(
        name="LNAME-1",
        general_modifier=["SOME ASCII TEXT"],
        quantity="SOME ASCII TEXT",
        quantity_modifier=["SOME ASCII TEXT"],
        altered_form="SOME ASCII TEXT",
        entity="SOME ASCII TEXT",
        entity_modifier=["SOME ASCII TEXT"],
        entity_number="SOME ASCII TEXT",
        entity_part="SOME ASCII TEXT",
        entity_part_number="SOME ASCII TEXT",
        generic_source="SOME ASCII TEXT",
        source_part=["SOME ASCII TEXT"],
        source_part_number=["SOME ASCII TEXT"],
        conditions=["SOME ASCII TEXT"],
        standard_symbol="SOME ASCII TEXT",
        private_symbol="SOME ASCII TEXT",
    )

    return ln


def _add_groups(
    lf: LogicalFile,
    channels: tuple[eflr_types.ChannelItem, ...],
    processes: tuple[eflr_types.ProcessItem, ...],
) -> tuple[eflr_types.GroupItem, ...]:
    g1 = lf.add_group(
        name="ChannelGroup",
        description="Group of channels",
        object_list=[channels[1], channels[2]],
    )

    g2 = lf.add_group(
        name="ProcessGroup",
        description="Group of processes",
        object_list=[processes[0], processes[1]],
    )

    g3 = lf.add_group(
        name="MultiGroup", description="Group of groups", group_list=[g1, g2]
    )

    g4 = lf.add_group(
        name="Mixed-group",
        description="Mixed objects",
        object_list=[channels[4], channels[5], processes[1]],
        group_list=[g1, g3],
    )

    return g1, g2, g3, g4


def create_dlis_file_object() -> DLISFile:
    df = DLISFile(storage_unit_label=make_sul())

    lf = df.add_logical_file(file_header=make_file_header())

    _add_origin(lf)

    axes = _add_axes(lf)
    ln = _add_long_name(lf)
    channels = _add_channels(lf, axes[0], ln)
    frame = _add_frame(lf, *channels[4:9])
    zones = _add_zones(lf)
    params = _add_parameters(lf, zones, ln)
    equipment = _add_equipment(lf)
    tools = _add_tools(lf, equipment, params, channels)
    computations = _add_computation(lf, axes, zones, tools, ln)
    processes = _add_processes(lf, params, channels, computations)
    _add_splices(lf, channels, zones)
    _add_calibrations(lf, axes, channels, params)
    wrp = _add_well_reference_points(lf)
    _add_paths(lf, frame, wrp[0], channels)
    _add_messages(lf)
    _add_comments(lf)
    _add_no_formats(lf)
    _add_groups(lf, channels, processes)

    return df


def write_short_dlis(
    fname: Union[str, os.PathLike[str]], data: Union[dict, os.PathLike[str], np.ndarray]
) -> None:
    df = create_dlis_file_object()
    df.write(fname, data=data)



import logging
from typing import Optional, Generator

from dliswriter.logical_record.core.logical_record.segment_attributes import SegmentAttributes
from dliswriter.utils.internal.internal_enums import RepresentationCode as RepC


logger = logging.getLogger(__name__)


class LogicalRecordBytes:
    """Wrap bytes of a LogicalRecord and StorageUnitLabel, adding segmenting functionalities."""

    padding: bytes = RepC.USHORT.convert(1)  #: padding byte added if the number of bytes in a segment is odd

    def __init__(self, bts: bytes, lr_type_struct: bytes, is_eflr: bool = False):
        """Initialise a LogicalRecordBytes object.

        Args:
            bts             :   Full bytes describing a logical record.
            lr_type_struct  :   Bytes describing the type of the logical record.
            is_eflr         :   True if the bytes describe an explicitly formatted logical record, False otherwise.
        """

        self._bts = bts
        self._size = len(bts)

        self._lr_type_struct = lr_type_struct
        self._is_eflr = is_eflr

    @property
    def bts(self) -> bytes:
        """Bytes describing a logical record."""

        return self._bts

    @property
    def size(self) -> int:
        """Number of bytes in the logical record description."""

        return self._size

    def make_segment(self, start_pos: int = 0, n_bytes: Optional[int] = None) -> tuple[bytes, int]:
        """Create a segment of the logical record bytes.

        Logical record bytes are often too long to fit into a visible record (physical unit of a DLIS file).
        In such case, they have to be split into several segments. Each segment should have a header which includes
        metadata such as segment size, whether this is the first/last segment of this logical record, what is the type
        of the logical record, etc.

        Args:
            start_pos   :   Starting index of the segment.
            n_bytes     :   Number of bytes - following the starting index - to be put in the segment. If not specified,
                            all bytes from start_pos will be included.

        Returns:
            2-tuple of:
                bytes   :   Bytes of the logical record segment, including the header with metadata.
                int     :   Total size of the segment (in bytes).
        """

        if n_bytes is None:  # include all bytes from start_pos till the end
            n_bytes = self._size - start_pos
            end_pos = None
            is_last = True
        else:
            end_pos = start_pos + n_bytes
            if end_pos > self._size:
                raise ValueError("Logical record too short for the requested bytes")
            is_last = end_pos == self._size

        if n_bytes < 12:
            raise ValueError(f"Logical Record segment body cannot be shorter than 12 bytes (got {n_bytes})")

        segment_attributes = SegmentAttributes(
            is_eflr=self._is_eflr,
            is_first=(start_pos == 0),
            is_last=is_last
        )

        size = n_bytes + 4  # adding header size - 4 bytes
        if size % 2:
            # total segment size must be even; if the number of bytes is odd, add a padding byte
            size += 1
            segment_attributes.has_padding = True

        header_bytes = RepC.UNORM.convert(size) + segment_attributes.to_struct() + self._lr_type_struct

        new_bts = header_bytes + self._bts[start_pos:end_pos]
        if segment_attributes.has_padding:
            new_bts += self.padding  # add the promised padding byte

        return new_bts, size

    def make_segments(self, max_n_bytes: int) -> Generator:
        """Define a generator which splits the logical record bytes into segments of given maximal size.

        See make_segment for a more detailed description.

        Args:
            max_n_bytes :   Maximal number of bytes in a segment body. Note that 4 more bytes will be added for the
                            header (and one more byte for the padding if the number of bytes is odd).
                            Must be an integer of minimal value of 12. Only this condition (>= 12) is checked;
                            other checks (type etc.) are left out for performance reasons.

        Yields:
            bytes   :   Bytes of a logical record segment, including an added header.
        """

        start_pos = 0  # start from the beginning of the logical record bytes
        remaining_size = self._size  # all bytes will be processed; self._size is assumed to always be >=12

        if max_n_bytes < 24:
            # minimal length of a logical record segment is 16 (of which 4 bytes are reserved for header),
            # so for the splitting to work correctly max_n_bytes must be >= 24, which is twice the min segment body size
            raise ValueError(f"Max size of a logical record segment body cannot be less than 24 (got {max_n_bytes})")

        while remaining_size > 0:
            n_bytes = min(remaining_size, max_n_bytes)  # size of the current (to be created) segment body
            future_remaining_size = remaining_size - n_bytes  # how many bytes will be left for a next segment
            if 0 < future_remaining_size < 12:
                # the next segment would be shorter than 12 bytes, and this is not allowed
                # so the current segment must be shortened by a few bytes which will be passed over to the next segment;
                n_bytes -= (12 - future_remaining_size)
                future_remaining_size = 12

            yield self.make_segment(start_pos, n_bytes)

            # update values before next iteration of the loop
            remaining_size = future_remaining_size
            start_pos += n_bytes



import pytest

from dliswriter.logical_record.eflr_types.tool import ToolSet, ToolItem


@pytest.mark.parametrize(("name", "description", "status", "param_names", "channel_names"), (
        ("TOOL-1", "SOME TOOL", 1, ["param3", "param2"], ["channel1", "channel2"]),
        ("Tool-X", "desc", 0, ["param1"], ["chan"])
))
def test_creating_tool(name: str, description: str, status: int, param_names: list[str], channel_names: list[str],
                       request: pytest.FixtureRequest) -> None:
    """Test creating ToolItem."""

    tool = ToolItem(
        name,
        description=description,
        status=status,
        parameters=[request.getfixturevalue(v) for v in param_names],
        channels=[request.getfixturevalue(v) for v in channel_names],
        parent=ToolSet()
    )

    assert tool.name == name
    assert tool.description.value == description
    assert tool.status.value == status

    for i, pn in enumerate(param_names):
        assert tool.parameters.value[i] is request.getfixturevalue(pn)

    for i, cn in enumerate(channel_names):
        assert tool.channels.value[i] is request.getfixturevalue(cn)

    assert isinstance(tool.parent, ToolSet)
    assert tool.parent.set_name is None



import pytest
from pathlib import Path
import os
from typing import Generator
from dlisio import dlis  # type: ignore  # untyped library
import numpy as np

from dliswriter.file.file import DLISFile

from tests.common import load_dlis
from tests.dlis_files_for_testing.double_frame_dlis import write_double_frame_dlis


@pytest.fixture(scope="session")
def double_frame_data() -> tuple[dict, dict]:
    n_rows_1 = 100
    n_rows_2 = 200

    frame1_data = {
        "DEPTH": np.arange(start=0, stop=n_rows_1, dtype=np.uint32),
        "RPM": (10 * np.random.rand(n_rows_1)).astype(np.float32),
        "AMPLITUDE": np.random.rand(n_rows_1, 10),
    }

    frame2_data = {
        "DEPTH": (np.arange(start=0, stop=n_rows_2) / 10).astype(np.uint32),
        "RPM": (np.arange(n_rows_2) % 10).astype(np.float32),
        "AMPLITUDE": (np.arange(n_rows_2 * 5).reshape(n_rows_2, 5) % 6).astype(np.float32),
    }

    return frame1_data, frame2_data


@pytest.fixture(scope="session")
def double_frame_dlis_path(base_data_path: Path) -> Generator:
    p = base_data_path / "double_frame.DLIS"
    yield p

    if p.exists():
        os.remove(p)


@pytest.fixture(scope="session")
def double_frame_dlis(
    double_frame_dlis_path: Path, double_frame_data: tuple[dict, dict]
) -> Generator:
    df = write_double_frame_dlis(double_frame_dlis_path, *double_frame_data)
    yield df


@pytest.fixture(scope="session")
def double_frame_dlis_contents(
    double_frame_dlis_path: Path, double_frame_dlis: DLISFile
) -> dlis.LogicalFile:
    with load_dlis(double_frame_dlis_path) as f:
        yield f


@pytest.fixture(scope="session")
def channels(double_frame_dlis_contents: dlis.LogicalFile) -> list[dlis.Channel]:
    return double_frame_dlis_contents.channels  # type: ignore  # return type not recognised


def test_channel_names(channels: list[dlis.Channel]) -> None:
    assert len(channels) == 6

    names = [c.name for c in channels]
    assert names.count("DEPTH") == 2
    assert names.count("RPM") == 2
    assert names.count("AMPLITUDE") == 2


def test_channel_copy_numbers(channels: list[dlis.Channel]) -> None:
    cn = [c.copynumber for c in channels]
    assert cn.count(0) == 3
    assert cn.count(1) == 3


@pytest.mark.parametrize("nr", (0, 1))
def test_frame_channels(double_frame_dlis_contents: dlis.LogicalFile, nr: int) -> None:
    frame = double_frame_dlis_contents.frames[nr]
    assert (
        frame.copynumber == 0
    )  # frames have different names, so can have the same copy number

    ch = frame.channels
    assert len(ch) == 3

    assert ch[0].name == "DEPTH"
    assert ch[1].name == "RPM"
    assert ch[2].name == "AMPLITUDE"

    for ch in ch:
        assert ch.copynumber == nr


@pytest.mark.parametrize("nr", (0, 1))
def test_frame_data(
    double_frame_dlis_contents: dlis.LogicalFile,
    double_frame_data: tuple[dict, dict],
    nr: int,
) -> None:

    frame_channels = double_frame_dlis_contents.frames[nr].channels
    data = double_frame_data[nr]

    assert (frame_channels[0].curves() == data["DEPTH"]).all()
    assert (frame_channels[1].curves() == data["RPM"]).all()
    assert (frame_channels[2].curves() == data["AMPLITUDE"]).all()


def test_dataset_names(double_frame_dlis: DLISFile) -> None:
    file_channels = double_frame_dlis.logical_files[0].channels

    assert file_channels[0].dataset_name == "DEPTH"
    assert file_channels[1].dataset_name == "RPM"
    assert file_channels[2].dataset_name == "AMPLITUDE"

    assert file_channels[3].dataset_name == "DEPTH__1"
    assert file_channels[4].dataset_name == "RPM__1"
    assert file_channels[5].dataset_name == "AMPLITUDE__1"

    assert all(
        dn in double_frame_dlis.logical_files[0]._data_dict
        for dn in ("DEPTH", "RPM", "AMPLITUDE", "DEPTH__1", "RPM__1", "AMPLITUDE__1")
    )


@pytest.mark.parametrize(
    ("channel_name", "dataset_name"), (("DEPTH", "depth"), ("XYZ", "XZY"))
)
def test_unique_dataset_name_from_dataset_name(
    double_frame_dlis: DLISFile, channel_name: str, dataset_name: str
) -> None:

    ch = double_frame_dlis.logical_files[0].add_channel(
        channel_name, dataset_name=dataset_name
    )
    assert ch.dataset_name == dataset_name


@pytest.mark.parametrize(
    ("channel_name", "dataset_name"), (("RPM", "RPM"), ("new_channel", "AMPLITUDE"))
)
def test_unique_dataset_name_but_dataset_name_exists(
    double_frame_dlis: DLISFile, channel_name: str, dataset_name: str
) -> None:

    with pytest.raises(
        ValueError, match=f"A data set with name '{dataset_name}' already exists"
    ):
        double_frame_dlis.logical_files[0].add_channel(
            channel_name, dataset_name=dataset_name
        )


@pytest.mark.parametrize("channel_name", ("NEW_CHANNEL", "XXX"))
def test_unique_dataset_name_from_new_channel_name(
    double_frame_dlis: DLISFile, channel_name: str
) -> None:

    ch = double_frame_dlis.logical_files[0].add_channel(channel_name)
    assert ch.dataset_name == channel_name


@pytest.mark.parametrize("channel_name", ("ABC", "SURFACE_RPM"))
def test_unique_dataset_name_from_repeated_channel_name(
    double_frame_dlis: DLISFile, channel_name: str
) -> None:
    chs = []
    for i in range(5):
        chs.append(double_frame_dlis.logical_files[0].add_channel(channel_name))

    assert chs[0].name == channel_name
    assert chs[0].dataset_name == channel_name

    for i in range(1, 5):
        assert chs[i].name == channel_name
        assert chs[i].dataset_name == f"{channel_name}__{i}"

    assert len(set(chs)) == 5  # 5 separate objects


def test_multiple_axes(double_frame_dlis_contents: dlis.LogicalFile) -> None:
    frames = double_frame_dlis_contents.frames
    for i in range(2):
        axes = frames[i].channels[0].axis
        assert len(axes) == 1
        axis = axes[0]
        assert axis.name == "AXIS"
        assert axis.copynumber == i

    assert (
        frames[0].channels[0].axis[0] is not frames[1].channels[0].axis[0]
    )  # separate objects



import pytest

from dliswriter.logical_record.eflr_types.well_reference_point import WellReferencePointSet, WellReferencePointItem


@pytest.mark.parametrize(("name", "v_zero", "m_decl", "c1_name", "c1_value", "c2_name", "c2_value"), (
        ("AQLN WELL-REF", "AQLN vertical_zero", 999.51, "Latitude", 40.395240, "Longitude", 27.792470),
        ("WRP-X", "vz20", 112.3, "X", 20, "Y", -0.3)
))
def test_from_config(name: str, v_zero: str, m_decl: float, c1_name: str, c1_value: float, c2_name: str,
                     c2_value: float) -> None:
    """Test creating WellReferencePointItem."""

    w = WellReferencePointItem(
        name,
        vertical_zero=v_zero,
        magnetic_declination=m_decl,
        coordinate_1_name=c1_name,
        coordinate_1_value=c1_value,
        coordinate_2_name=c2_name,
        coordinate_2_value=c2_value,
        parent=WellReferencePointSet()
    )

    assert w.name == name
    assert w.vertical_zero.value == v_zero
    assert w.magnetic_declination.value == m_decl
    assert w.coordinate_1_name.value == c1_name
    assert w.coordinate_1_value.value == c1_value
    assert w.coordinate_2_name.value == c2_name
    assert w.coordinate_2_value.value == c2_value

    assert isinstance(w.parent, WellReferencePointSet)
    assert w.parent.set_name is None



from datetime import datetime
import numpy as np
from typing import Callable, Any, Iterable, Union, Optional

from dliswriter.utils.internal.internal_enums import RepresentationCode
from dliswriter.utils.internal.types import numpy_dtype_type


def get_ascii_bytes(value: str, required_length: int, justify_left: bool = False) -> bytes:
    """Encode a string value as ASCII.

    Args:
        value           :   Value to convert.
        required_length :   Required length of the string. Padding will be added if the original string is shorter.
        justify_left    :   If True, pad with trailing blanks. Otherwise, pad with preceding blanks.

    Returns:
        ASCII encoded bytes.
    """

    lv = len(value)
    if lv > required_length:
        raise ValueError(f"Provided string is too long ({lv} chars) for required length of {required_length}")

    padding = (required_length - lv) * ' '

    if justify_left:
        padded_value = value + padding
    else:
        padded_value = padding + value

    return padded_value.encode('ascii')


def _filter_codes(cond: Callable) -> tuple[RepresentationCode, ...]:
    """Return a tuple containing all representation codes fulfilling a certain condition ('cond')."""

    return tuple(code for code in RepresentationCode.__members__.values() if cond(code))


class ReprCodeConverter:
    """Choose a representation code based on the provided data or determine data type from a representation code."""

    class ReprCodeError(ValueError):
        """Error raised if a representation code for a given value cannot be determined."""

        pass

    float_codes = _filter_codes(lambda code: code.value <= 11)          #: representation codes for floats
    sint_codes = _filter_codes(lambda code: 12 <= code.value <= 14)     #: representation codes for signed ints
    uint_codes = _filter_codes(lambda code: 15 <= code.value <= 18)     #: representation codes for unsigned ints
    int_codes = sint_codes + uint_codes                                 #: representation codes for all integers
    numeric_codes = float_codes + int_codes                             #: representation codes for all numbers

    # mapping of numpy dtype names on corresponding representation codes
    numpy_dtypes_to_repr_codes: dict[str, RepresentationCode] = {
        'int8': RepresentationCode.SSHORT,
        'int16': RepresentationCode.SNORM,
        'int32': RepresentationCode.SLONG,
        'uint8': RepresentationCode.USHORT,
        'uint16': RepresentationCode.UNORM,
        'uint32': RepresentationCode.ULONG,
        'float32': RepresentationCode.FSINGL,
        'float64': RepresentationCode.FDOUBL
    }

    # mapping of numerical representation codes on corresponding numpy dtypes
    repr_codes_to_numpy_dtypes: dict[RepresentationCode, numpy_dtype_type]\
        = {v: getattr(np, k) for k, v in numpy_dtypes_to_repr_codes.items()}

    # mapping of different object types on corresponding representation codes
    generic_types: dict[type, RepresentationCode] = {
        datetime: RepresentationCode.DTIME,
        int: RepresentationCode.SLONG,
        float: RepresentationCode.FDOUBL,
        str: RepresentationCode.ASCII
    }

    @classmethod
    def validate_numpy_dtype(cls, number_type: numpy_dtype_type) -> tuple[str, RepresentationCode]:
        """Check that the provided value is a numpy dtype. Return the dtype name and the corresponding repr code."""

        if isinstance(number_type, np.dtype):
            number_type_name = number_type.name
        elif isinstance(number_type, type) and issubclass(number_type, np.generic):
            number_type_name = number_type.__name__
        else:
            raise ValueError(f"{number_type} is not a numpy dtype")
        if number_type_name not in ReprCodeConverter.numpy_dtypes_to_repr_codes:
            raise ValueError(f"Dtype {number_type_name} is not supported; "
                             f"allowed dtypes are: {', '.join(ReprCodeConverter.numpy_dtypes_to_repr_codes)}")

        return number_type_name, cls.numpy_dtypes_to_repr_codes[number_type_name]

    @classmethod
    def determine_repr_code_from_numpy_dtype(cls, dt: numpy_dtype_type) -> RepresentationCode:
        """Determine representation code for a given numpy dtype."""

        return cls.validate_numpy_dtype(dt)[1]

    @classmethod
    def determine_repr_code_from_generic_type(cls, t: type) -> RepresentationCode:
        """Determine representation code for a given type (e.g. int, float, str, etc.)."""

        if not isinstance(t, type):
            raise TypeError(f"{t} is not a type")

        repr_code = cls.generic_types.get(t, None)
        if not repr_code:
            raise cls.ReprCodeError(f"Cannot determine representation code for type {t}")
        return repr_code

    @classmethod
    def _determine_repr_code_single(cls, value: Any) -> RepresentationCode:
        """Determine representation code for a value (which is not a list/tuple etc., but might be a numpy array)."""

        if isinstance(value, (np.generic, np.ndarray)):
            return cls.determine_repr_code_from_numpy_dtype(value.dtype)
        return cls.determine_repr_code_from_generic_type(type(value))

    @classmethod
    def _determine_repr_code_multiple(cls, values: Iterable[Any]) -> RepresentationCode:
        """Determine representation code for an iterable of values.

        Note:
            A single representation code for all values is returned. In case it is not the same representation code
            for all values, the method will try to find one that fits all the values (e.g. SLONG if some values are
            SLONG and some SNORM, od FDOUBL if all values are numbers). If that's not possible, a ReprCodeError
            is raised.
        """

        repr_codes = [cls._determine_repr_code_single(v) for v in values]
        if len(set(repr_codes)) == 1:
            return repr_codes[0]

        if not all(rc in cls.numeric_codes for rc in repr_codes):
            raise cls.ReprCodeError(f"Cannot determine a common representation code for values: {values} "
                                    f"(proposed representation codes are: {repr_codes})")

        # at this stage we know all codes are numeric

        if any(rc in cls.float_codes for rc in repr_codes):
            # if any of them is a float - return the float code
            return RepresentationCode.FDOUBL

        if any(all(rc in codes for rc in repr_codes) for codes in (cls.float_codes, cls.sint_codes, cls.uint_codes)):
            # only floats, only signed ints, or only unsigned ints
            return max(repr_codes)

        if any(rc in cls.sint_codes for rc in repr_codes):
            # if any of them is a signed int - return a signed int code
            return RepresentationCode.SLONG

        raise cls.ReprCodeError(f"Cannot determine a representation code for values: {values}")

    @classmethod
    def determine_repr_code_from_value(cls, v: Any) -> RepresentationCode:
        """Determine representation code for a value.

        In case the value is a list or a tuple, the method will try to find a single representation code which fits
        all elements of the value.
        """

        if isinstance(v, (list, tuple)):
            return cls._determine_repr_code_multiple(v)
        return cls._determine_repr_code_single(v)

    @staticmethod
    def determine_numpy_dtype_from_repr_code(repr_code: Union[RepresentationCode, None],
                                             default: Optional[RepresentationCode] = None) -> numpy_dtype_type:
        """Determine a numpy dtype for a given representation code.

        Args:
            repr_code   :   Representation code to convert to a numpy dtype.
            default     :   If not None and 'repr_code' is None, dtype corresponding to this repr code.

        Returns:
            A numpy.dtype object corresponding to the given representation code.
        """

        if repr_code is None:
            if default is not None:
                return ReprCodeConverter.determine_numpy_dtype_from_repr_code(default)
            else:
                raise ValueError("Expected a RepresentationCode; got None")

        return ReprCodeConverter.repr_codes_to_numpy_dtypes[repr_code]



from datetime import datetime, timezone
from typing import Any, TYPE_CHECKING
from functools import lru_cache

from dliswriter.utils.internal.internal_enums import RepresentationCode

if TYPE_CHECKING:
    from dliswriter.logical_record.core.eflr import EFLRItem


# offsets used in writing structs for UVARI representation code (see '_write_struct_uvari' function)
UNORM_OFFSET = 32768        #: offset added to values packed as UNORM; '10' and 14 zeros
ULONG_OFFSET = 3221225472   #: offset added to values packed as ULONG; 11 and 30 zeros


def write_struct_dtime(date_time: datetime) -> bytes:
    """Convert a datetime object to bytes according to the RP66 V1 standard.

    The date-time is converted to UTC, which is marked with the corresponding time zone code (2, "GMT").

    Args:
        date_time: The datetime object to be converted.

    Returns:
        Bytes representing the date time.

    From RP66:
        Y = Years Since 1900 (Range 0 to 255)
        TZ = Time Zone (0 = Local Standard, 1 = Local Daylight Savings, 2 = Greenwich Mean Time)
        M = Month of the Year (Range 1 to 12)
        D = Day of Month (Range 1 to 31)
        H = Hours Since Midnight (Range 0 to 23)
        MN = Minutes Past Hour (Range 0 to 59)
        S = Seconds Past Minute (Range 0 to 59)
        MS = Milliseconds Past Second (Range 0 to 999)

        9:20:15.62 PM, April 19, 1987 (DST) =
        87 years since 1900, 4th month, 19th day,
        21 hours since midnight, 20 minutes past hour,
        15 seconds past minute, 620 milliseconds past second =
            01010111 00010100 00010011 00010101
            00010100 00001111 00000010 01101100

        Representation codes for the different parts:
            - milliseconds: UNORM
            - timezone & month: 4-bit unsigned integer (doesn't have a proper representation code in RP66)
            - other parts: USHORT

    """

    # bring the date-time to UTC
    date_time = date_time.astimezone(timezone.utc)

    value = b''

    time_zone = '{0:04b}'.format(2)  # date-time converted to UTC (previously called GMT), so we use the GMT code
    month = '{0:04b}'.format(date_time.month)

    value += RepresentationCode.USHORT.convert(date_time.year - 1900)
    value += RepresentationCode.USHORT.convert(int(time_zone + month, 2))
    value += RepresentationCode.USHORT.convert(date_time.day)
    value += RepresentationCode.USHORT.convert(date_time.hour)
    value += RepresentationCode.USHORT.convert(date_time.minute)
    value += RepresentationCode.USHORT.convert(date_time.second)
    value += RepresentationCode.UNORM.convert(min(round(date_time.microsecond / 1000), 999))

    return value


def write_struct_ascii(value: Any) -> bytes:
    """Convert value to str, encode as ASCII, and represent as bytes.

    The first bytes are the number of characters in the value (converted to str).
    """

    value_str = str(value)
    return write_struct_uvari(len(value_str)) + value_str.encode('ascii')


def write_struct_uvari(value: int) -> bytes:
    """Convert an integer to bytes. The format (USHORT/UNORM/ULONG) is chosen depending on the provided value."""

    if value < 128:
        return RepresentationCode.USHORT.convert(value)

    if value < 16384:
        return RepresentationCode.UNORM.convert(value + UNORM_OFFSET)

    # >= 16384
    return RepresentationCode.ULONG.convert(value + ULONG_OFFSET)


def write_struct_obname(value: "EFLRItem") -> bytes:
    """Create a reference to an EFLRObject, based on the object's name."""

    if value.origin_reference is None:
        raise RuntimeError(f"Origin reference of {value} has not been specified")

    try:
        origin_reference = write_struct_uvari(value.origin_reference)
        copy_number = RepresentationCode.USHORT.convert(value.copy_number)
        name = write_struct_ascii(value.name)

        obname = origin_reference + copy_number + name

    except AttributeError:
        raise TypeError(f"'OBNAME' struct can only be written for an EFLR object; got {type(value)}: {value}")

    return obname


def write_struct_objref(value: "EFLRItem") -> bytes:
    """Create a reference to an EFLRObject, based on the object's name and set type it belongs to."""

    return write_struct_ascii(value.parent.set_type) + value.obname


def write_struct_status(value: int) -> bytes:
    """Represent status (1 or 0) as bytes."""

    if value != 0 and value != 1:
        raise ValueError(f"STATUS must be 1 (meaning ALLOWED/TRUE/ON) or 0 (meaning DISALLOWED/FALSE/OFF); got {value}")

    return RepresentationCode.USHORT.convert(value)


# dictionary collecting all the individual write_struct sub-functions for faster access in the main function below
_struct_dict = {
    RepresentationCode.ASCII: write_struct_ascii,
    RepresentationCode.UVARI: write_struct_uvari,
    RepresentationCode.IDENT: write_struct_ascii,
    RepresentationCode.DTIME: write_struct_dtime,
    RepresentationCode.OBNAME: write_struct_obname,
    RepresentationCode.OBJREF: write_struct_objref,
    RepresentationCode.STATUS: write_struct_status
}


@lru_cache(maxsize=65536)
def write_struct(representation_code: RepresentationCode, value: Any) -> bytes:
    """Convert a value to bytes according to the RP66 V1 spec.

    Args:
        representation_code :   The way the value should be represented as.
        value               :   Value to be converted.

    Returns:
        Value converted to bytes depending on representation_code and RP66 V1 spec.
    """

    func = _struct_dict.get(representation_code, None)  # get a converter corresponding to the repr code
    if func:
        return func(value)  # type: ignore  # that's the point, we're calling for any type

    return representation_code.convert(value)  # if no converter was found, use the one built in the enum



import pytest
from datetime import datetime, timedelta
from typing import Any, Generator

from dliswriter.logical_record.core.attribute import Attribute, DTimeAttribute


@pytest.fixture
def attr() -> Generator:
    """Basic mock attribute for tests."""

    yield Attribute('some_attribute')


@pytest.mark.parametrize(("val", "unit"), (("m", 'm'), ("in", 'in')))
def test_setting_unit(attr: Attribute, val: str, unit: str) -> None:
    """Test that attribute units are set correctly."""

    attr.units = val
    assert attr.units is unit


def test_clearing_unit(attr: Attribute) -> None:
    """Test that removing a previously defined unit works correctly."""

    attr.units = None
    assert attr.units is None


@pytest.mark.parametrize("dts", ("2050/03/21 15:30:00", "2003.12.31 09:30:00"))
def test_parse_dtime(dts: str) -> None:
    """Check that correctly formatted date-time strings are parsed correctly."""

    dt = DTimeAttribute.parse_dtime(dts)
    assert isinstance(dt, datetime)


@pytest.mark.parametrize("dts", ("2050/21/03 15:30:00", "2050.03.02 15:30"))
def test_parse_dtime_wrong_format(dts: str) -> None:
    """Check that a ValueError is raised if a provided date-time string does not follow any of the allowed formats."""

    with pytest.raises(ValueError, match=".*does not conform to any of the allowed formats.*"):
        DTimeAttribute.parse_dtime(dts)


@pytest.mark.parametrize("dts", (object(), timedelta(seconds=12)))
def test_parse_dtime_wrong_type(dts: Any) -> None:
    """Check that a TypeError is raised if the provided date-time string is not actually a string."""

    with pytest.raises(TypeError, match="Expected a str.*"):
        DTimeAttribute.parse_dtime(dts)



import pytest
from typing import Any

from dliswriter.logical_record.eflr_types import EquipmentSet, EquipmentItem
from dliswriter import AttrSetup, high_compatibility_mode_decorator


@pytest.mark.parametrize(("name", "status", "serial_number"), (
        ("EQ1", 1, "9101-21391"),
        ("EQ2", 0, "5559101-21391"),
        ("EqX", 1, "12311")
))
def test_creation(name: str, status: int, serial_number: str) -> None:
    """Check that EquipmentObject instances are correctly created from config."""

    eq = EquipmentItem(
        name,
        status=status,
        serial_number=serial_number,
        parent=EquipmentSet()
    )

    assert eq.name == name
    assert eq.status.value == status
    assert isinstance(eq.status.value, int)
    assert eq.serial_number.value == serial_number
    assert isinstance(eq.serial_number.value, str)

    assert isinstance(eq.parent, EquipmentSet)
    assert eq.parent.set_name is None


def test_params_and_units() -> None:
    """Check setting up EquipmentObject's parameters and units."""

    eq = EquipmentItem(
        "Equipment-1",
        **{
            'height': {'value': 140, 'units': 'in'},
            'length': AttrSetup(230.78, 'cm'),
            'minimum_diameter': AttrSetup(2.3, units='m'),
            'maximum_diameter': AttrSetup(value=3.2, units='m'),
            'weight': {'value': 1.2, 'units': 't'},
            'hole_size': {'value': 323.2, 'units': 'm'},
            'pressure': {'value': 18000, 'units': 'psi'},
            'temperature': {'value': 24, 'units': 'degC'},
            'vertical_depth': {'value': 587, 'units': 'm'},
            'radial_drift': AttrSetup(23.22, 'm'),
            'angular_drift': AttrSetup(32.5, 'm')
        },
        parent=EquipmentSet()
    )

    def check(name: str, val: Any, unit: str) -> None:
        attr = getattr(eq, name)
        assert attr.value == val
        assert attr.units == unit

    check('height', 140, 'in')
    check('length', 230.78, 'cm')
    check('minimum_diameter', 2.3, 'm')
    check('maximum_diameter', 3.2, 'm')
    check('weight', 1.2, 't')
    check('hole_size', 323.2, 'm')
    check('pressure', 18000, 'psi')
    check('temperature', 24, 'degC')
    check('vertical_depth', 587, 'm')
    check('radial_drift', 23.22, 'm')
    check('angular_drift', 32.5, 'm')


@pytest.mark.parametrize("sn", ("124-111", "123_B3", "A12"))
@high_compatibility_mode_decorator
def test_serial_number_compatible(sn: str) -> None:
    EquipmentItem("EQUIPMENT-1", serial_number=sn, parent=EquipmentSet())


@pytest.mark.parametrize("sn", ("123.1", "112 131", "213/1"))
@high_compatibility_mode_decorator
def test_type_not_compatible(sn: str) -> None:
    with pytest.raises(ValueError, match=".*strings can contain only uppercase characters, digits, dashes, .*"):
        EquipmentItem("EQUIPMENT-1", serial_number=sn, parent=EquipmentSet())


@pytest.mark.parametrize("name", ("EQ-1", "EQP3", "11C"))
@high_compatibility_mode_decorator
def test_name_compatible(name: str) -> None:
    EquipmentItem(name, parent=EquipmentSet())


@pytest.mark.parametrize("name", ("EQ 1", "Equip-1", "EQ.32"))
@high_compatibility_mode_decorator
def test_label_not_compatible(name: str) -> None:
    with pytest.raises(ValueError, match=".*strings can contain only uppercase characters, digits, dashes, .*"):
        EquipmentItem(name, parent=EquipmentSet())



import pytest
from dlisio import dlis    # type: ignore  # untyped library
from typing import Union

from tests.common import check_list_of_objects


def test_computation(short_dlis: dlis.file.LogicalFile) -> None:
    """Check that the number of Computation objects in the DLIS file matches the expected one."""

    comps = short_dlis.computations
    assert len(comps) == 3


@pytest.mark.parametrize(("idx", "name", "properties", "zone_names", "axis_name", "values"), (
        (0, "COMPT-1", ["LOCALLY-DEFINED", "AVERAGED"], ["Zone-1", "Zone-2"], "Axis-3", [[100, 200, 300], [1, 2, 3]]),
        (1, "COMPT2", ["UNDER-SAMPLED", "AVERAGED"], ["Zone-1", "Zone-3"], "Axis-4", [[1.5, 2.5], [4.5, 3.2]]),
        (2, "COMPT-X", ["OVER-SAMPLED"], [], "Axis-1", [list(range(12, 24))]),
))
def test_computation_params(short_dlis: dlis.file.LogicalFile, idx: int, name: str, properties: list[str],
                            zone_names: list[str], axis_name: str, values: list[Union[int, float]]) -> None:
    """Check attributes of Computation objects in the new DLIS file."""

    comp = short_dlis.computations[idx]

    assert comp.name == name
    assert comp.properties == properties
    assert comp.axis[0].name == axis_name
    assert comp.values.tolist() == values
    assert comp.origin == 42

    check_list_of_objects(comp.zones, zone_names)


def test_computation_long_name(short_dlis: dlis.file.LogicalFile) -> None:
    assert short_dlis.computations[0].long_name == "COMPT1"
    assert short_dlis.computations[1].long_name is short_dlis.longnames[0]
    assert short_dlis.computations[2].long_name == "Computation not added to process"



from dlisio import dlis    # type: ignore  # untyped library


def test_storage_unit_label(short_dlis: dlis.file.LogicalFile) -> None:
    """Check attributes of Storage Unit Label in the new DLIS file."""

    sul = short_dlis.storage_label()
    assert sul['id'].rstrip(' ') == "DEFAULT STORAGE SET"
    assert sul['sequence'] == 1
    assert sul['maxlen'] == 8192



import logging
from typing import Any

from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem, DimensionedItem
from dliswriter.utils.internal.internal_enums import EFLRType, RepresentationCode as RepC
from dliswriter.utils.enums import CalibrationMeasurementPhase
from dliswriter.logical_record.eflr_types.axis import AxisSet
from dliswriter.logical_record.core.attribute import (EFLRAttribute, NumericAttribute,
                                                      DTimeAttribute, DimensionAttribute, IdentAttribute)
from dliswriter.utils.internal.value_checkers import validate_string


logger = logging.getLogger(__name__)


class CalibrationMeasurementItem(EFLRItem, DimensionedItem):
    """Model an object being part of CalibrationMeasurement EFLR."""

    parent: "CalibrationMeasurementSet"

    def __init__(self, name: str, parent: "CalibrationMeasurementSet", **kwargs: Any) -> None:
        """Initialise CalibrationMeasurementItem.

        Args:
            name        :   Name of the CalibrationMeasurementItem.
            parent      :   Parent CalibrationMeasurementSet of this CalibrationMeasurementItem.
            **kwargs    :   Values of to be set as characteristics of the CalibrationMeasurementItem Attributes.
        """

        self.phase = IdentAttribute(
            'phase',
            converter=CalibrationMeasurementPhase.make_converter('phases', allow_none=True)
        )
        self.measurement_source = EFLRAttribute(
            'measurement_source', representation_code=RepC.OBJREF, object_class=EFLRSet)
        self.type = IdentAttribute('type', converter=validate_string)
        self.dimension = DimensionAttribute('dimension')
        self.axis = EFLRAttribute('axis', object_class=AxisSet, multivalued=True)
        self.measurement = NumericAttribute('measurement', multivalued=True, multidimensional=True)
        self.sample_count = NumericAttribute('sample_count', int_only=True)
        self.maximum_deviation = NumericAttribute('maximum_deviation', multivalued=True, multidimensional=True)
        self.standard_deviation = NumericAttribute('standard_deviation', multivalued=True, multidimensional=True)
        self.begin_time = DTimeAttribute('begin_time', allow_float=True)
        self.duration = NumericAttribute('duration')
        self.reference = NumericAttribute('reference', multivalued=True, multidimensional=True)
        self.standard = NumericAttribute('standard', multivalued=True, multidimensional=True)
        self.plus_tolerance = NumericAttribute('plus_tolerance', multivalued=True, multidimensional=True)
        self.minus_tolerance = NumericAttribute('minus_tolerance', multivalued=True, multidimensional=True)

        super().__init__(name, parent=parent, **kwargs)

    def _run_checks_and_set_defaults(self) -> None:
        """Check that the number of max/std deviations, standards, and tolerances is equal."""

        self._check_axis_vs_dimension()

        controlled_attrs = (self.maximum_deviation, self.standard_deviation, self.standard,
                            self.plus_tolerance, self.minus_tolerance)

        value_counts = self.count_attributes(*controlled_attrs)

        if len(set(value_counts.values())) > 1:
            raise RuntimeError(f"Number of values all numeric attributes of Calibration Coefficient should be equal; "
                               f"got {', '.join('{} for {}'.format(v, k) for k, v in value_counts.items())}")

        for attr in controlled_attrs:
            self._check_or_set_value_dimensionality(attr.value, value_label=attr.label)


class CalibrationMeasurementSet(EFLRSet):
    """Model CalibrationMeasurement EFLR."""

    set_type = 'CALIBRATION-MEASUREMENT'
    logical_record_type = EFLRType.STATIC
    item_type = CalibrationMeasurementItem


CalibrationMeasurementItem.parent_eflr_class = CalibrationMeasurementSet



import pytest

from dliswriter.logical_record.eflr_types.axis import AxisItem, AxisSet
from dliswriter import high_compatibility_mode_decorator


@pytest.mark.parametrize(('name', 'axis_id', 'units'), (
    ('Axis-1', 'First axis', 'm'),
    ('Axis-X', 'Another axis', 's')
))
def test_str_attributes(name: str, axis_id: str, units: str) -> None:
    axis = AxisItem(name, axis_id=axis_id, spacing={'units': units}, parent=AxisSet())

    assert axis.name == name
    assert axis.axis_id.value == axis_id
    assert axis.spacing.units == units


@pytest.mark.parametrize(('given', 'expected'), (
        ([40.395241, 27.792471], [40.395241, 27.792471]),
        (12.121, [12.121]),
        ([8], [8])
))
def test_coordinates(given: list, expected: list) -> None:
    axis = AxisItem('some_name', coordinates=given, parent=AxisSet())
    assert axis.coordinates.value == expected


def test_copy_numbers() -> None:
    parent = AxisSet()

    for i in range(5):
        ax = AxisItem('XYZ', parent=parent)
        assert ax.copy_number == i


@pytest.mark.parametrize("name", ("AXIS-1", "TIME_AXIS", "123"))
@high_compatibility_mode_decorator
def test_name_compatible(name: str) -> None:
    AxisItem(name, parent=AxisSet())  # no error = name accepted, test passed


@pytest.mark.parametrize("name", ("Axis 3", "NEW AXIS", "1.2A"))
@high_compatibility_mode_decorator
def test_name_not_compatible(name: str) -> None:
    with pytest.raises(ValueError, match=".*strings can contain only uppercase characters, digits, dashes, .*"):
        AxisItem(name, parent=AxisSet())


@pytest.mark.parametrize("aid", ("SOME-AXIS", "AXIS_FOR_SPECIFIC_STUFF-124", "3112-1212"))
@high_compatibility_mode_decorator
def test_axis_id_compatible(aid: str) -> None:
    AxisItem("AX-1", axis_id=aid, parent=AxisSet())


@pytest.mark.parametrize("aid", ("The third axis", "AX FOR THINGS", "A.123"))
@high_compatibility_mode_decorator
def test_axis_id_not_compatible(aid: str) -> None:
    with pytest.raises(ValueError, match=".*strings can contain only uppercase characters, digits, dashes, .*"):
        AxisItem("AXIS3", axis_id=aid, parent=AxisSet())



import numpy as np
import pytest
import h5py  # type: ignore  # untyped library
from pathlib import Path
from typing import Union

from dliswriter.utils.source_data_wrappers import HDF5DataWrapper, SourceDataWrapper


@pytest.fixture(scope='session')
def mapping() -> dict:
    return {'time': '/contents/time', 'rad': '/contents/image0', 'amp': '/contents/image1', 'rpm': '/contents/rpm'}


def test_basic_properties(short_reference_data_path: Path, mapping: dict) -> None:
    w = HDF5DataWrapper(short_reference_data_path, mapping)
    assert isinstance(w.data_source, h5py.File)
    assert w.data_source.file

    assert w.n_rows == 100
    assert isinstance(w.dtype, np.dtype)
    assert isinstance(w.dtype.names, tuple)
    assert len(w.dtype.names) == 4


def test_creation(short_reference_data_path: Path, mapping: dict) -> None:
    w = HDF5DataWrapper(short_reference_data_path, mapping)
    assert w.dtype.names == ('time', 'rad', 'amp', 'rpm')
    assert w.dtype[0] == np.float64
    assert w.dtype[1] == (np.float64, (128,))
    assert w.dtype[2] == (np.float64, (128,))
    assert w.dtype[3] == np.float64


@pytest.mark.parametrize(('new_mapping', 'shapes'), (
        ({'time': 'contents/time', 'rpm': '/contents/rpm'}, [None, None]),
        ({'MD': '/contents/depth'}, [None]),
        ({'AMP': 'contents/image1', 'rpm': '/contents/rpm', 'time': 'contents/time'}, [(128,), None, None]),
        ({'AMP': 'contents/image1', 'RADIUS': '/contents/image0'}, [(128,), (128,)])
))
def test_alternative_mappings(short_reference_data_path: Path, new_mapping: dict, shapes: list) -> None:
    w = HDF5DataWrapper(short_reference_data_path, new_mapping)

    assert w.dtype.names == tuple(new_mapping.keys())
    for i in range(len(w.dtype.names)):
        assert w.dtype[i] == (np.float64 if shapes[i] is None else (np.float64, shapes[i]))


@pytest.mark.parametrize(('known_dtypes', 'dtype_check'), (
        ({'time': np.float32, 'rpm': np.float32}, (np.float32, np.float64, np.float64, np.float32)),
        ({'rpm': np.int32}, (np.float64, np.float64, np.float64, np.int32)),
        ({'rad': np.float32, 'amp': np.int32}, (np.float64, np.float32, np.int32, np.float64))
))
def test_creation_with_known_dtypes(short_reference_data_path: Path, known_dtypes: dict, dtype_check: tuple,
                                    mapping: dict) -> None:
    w = HDF5DataWrapper(short_reference_data_path, known_dtypes=known_dtypes, mapping=mapping)

    assert w.dtype[0] == dtype_check[0]
    assert w.dtype[1] == (dtype_check[1], (128,))
    assert w.dtype[2] == (dtype_check[2], (128,))
    assert w.dtype[3] == dtype_check[3]


def test_creation_from_superclass(short_reference_data_path: Path, mapping: dict) -> None:
    w = SourceDataWrapper.make_wrapper(short_reference_data_path, mapping=mapping)
    assert isinstance(w, HDF5DataWrapper)


@pytest.mark.parametrize(('from_idx', 'to_idx', 'n_rows'), ((0, None, 100), (0, 63, 63), (71, 88, 17)))
def test_creation_with_from_and_to_idx(short_reference_data_path: Path, mapping: dict, from_idx: int,
                                       to_idx: Union[int, None], n_rows: int) -> None:
    w = HDF5DataWrapper(short_reference_data_path, mapping=mapping, from_idx=from_idx, to_idx=to_idx)
    assert w.n_rows == n_rows


@pytest.mark.parametrize(('start', 'stop'), ((0, 20), (25, 30), (11, 12)))
def test_load_chunk_alternative_mapping(short_reference_data_path: Path, start: int, stop: int) -> None:
    w = HDF5DataWrapper(
        short_reference_data_path,
        mapping={'time': '/contents/time', 'rpm': '/contents/rpm', 'rad': '/contents/image0'}
    )

    chunk = w.load_chunk(start, stop)

    assert chunk.size == stop - start
    assert chunk.dtype.names == ('time', 'rpm', 'rad')
    assert chunk.dtype == w.dtype

    with h5py.File(short_reference_data_path, 'r') as data:
        assert (chunk['time'] == data['/contents/time'][start:stop]).all()
        assert (chunk['rpm'] == data['/contents/rpm'][start:stop]).all()
        assert (chunk['rad'] == data['/contents/image0'][start:stop]).all()


@pytest.mark.parametrize(("from_idx", "to_idx"), ((0, 12), (90, None)))
def test_getitem(short_reference_data_path: Path, mapping: dict, from_idx: int, to_idx: Union[int, None]) -> None:
    w = HDF5DataWrapper(short_reference_data_path, mapping=mapping, from_idx=from_idx, to_idx=to_idx)

    with h5py.File(short_reference_data_path, 'r') as data:
        for key in ('time', 'rpm', 'rad', 'amp'):
            assert isinstance(w[key], np.ndarray)
            assert (w[key] == data[mapping[key]][from_idx:to_idx]).all()



import os
from typing import Union

from dliswriter import (
    DLISFile,
    LogicalFile,
)
from dliswriter.utils.enums import FrameIndexType

from tests.dlis_files_for_testing.common import make_df


def _define_frame_from_data(lf: LogicalFile, name: str, data: dict) -> None:
    ax = lf.add_axis("AXIS")

    channels = []
    for i, (ch_name, ch_data) in enumerate(data.items()):
        ch = lf.add_channel(ch_name, data=ch_data)
        if not i:
            ch.axis.value = ax
        channels.append(ch)

    lf.add_frame(name, channels=channels, index_type=FrameIndexType.BOREHOLE_DEPTH)


def create_dlis_file_object(*data_dicts: dict) -> DLISFile:
    df = make_df()

    for i, d in enumerate(data_dicts):
        _define_frame_from_data(df.logical_files[0], f"FRAME{i+1}", d)

    return df


def write_double_frame_dlis(
    fname: Union[str, os.PathLike[str]], *frame_data: dict
) -> DLISFile:
    df = create_dlis_file_object(*frame_data)
    df.write(fname)

    return df



import pytest
import logging
import numpy as np

from dliswriter import DLISFile, high_compatibility_mode_decorator, enums


def _prepare_file_channel_not_in_frame() -> DLISFile:
    df = DLISFile()
    lf = df.add_logical_file()
    lf.add_origin("ORIGIN")
    ch1 = lf.add_channel("INDEX", data=np.arange(10))
    lf.add_channel("X", units="m", data=np.random.rand(10, 10))  # not added to frame
    ch3 = lf.add_channel("Y", data=np.arange(10, 20))
    lf.add_frame(
        "MAIN", channels=(ch1, ch3), index_type=enums.FrameIndexType.NON_STANDARD
    )

    return df


def test_channel_not_in_frame(caplog: pytest.LogCaptureFixture) -> None:

    with caplog.at_level(logging.WARNING, logger="dliswriter"):
        df = _prepare_file_channel_not_in_frame()
        df.logical_files[0].check_objects()
        assert "ChannelItem 'X' has not been added to any frame" in caplog.text


@high_compatibility_mode_decorator
def test_channel_not_in_frame_high_compat_mode() -> None:
    df = _prepare_file_channel_not_in_frame()

    with pytest.raises(
        RuntimeError, match="ChannelItem 'X' has not been added to any frame.*"
    ):
        df.logical_files[0].check_objects()


def _prepare_file_channel_in_multiple_frames() -> DLISFile:
    df = DLISFile()
    lf = df.add_logical_file()
    lf.add_origin("ORIGIN")
    ch_a = lf.add_channel("A")  # in 3 frames
    ch_b = lf.add_channel("B")
    ch_c = lf.add_channel("C")  # in 2 frames
    ch_d = lf.add_channel("D")

    lf.add_frame("MAIN", channels=(ch_a, ch_c))
    lf.add_frame("F2", channels=(ch_a, ch_b))
    lf.add_frame("F3", channels=(ch_a, ch_c, ch_d))

    return df


def test_channel_in_multiple_frames(caplog: pytest.LogCaptureFixture) -> None:

    with caplog.at_level(logging.WARNING, logger="dliswriter"):
        df = _prepare_file_channel_in_multiple_frames()
        df.logical_files[0].check_objects()
        assert "ChannelItem 'A' has been added to 3 frames" in caplog.text
        assert "ChannelItem 'C' has been added to 2 frames" in caplog.text


@high_compatibility_mode_decorator
def test_channel_in_multiple_frames_high_compat_mode() -> None:
    df = _prepare_file_channel_in_multiple_frames()

    with pytest.raises(
        RuntimeError, match="ChannelItem 'A' has been added to 3 frames.*"
    ):
        df.logical_files[0].check_objects()


def _prepare_file_with_sint_data() -> DLISFile:
    df = DLISFile()
    lf = df.add_logical_file()
    lf.add_origin("ORIGIN")
    ch1 = lf.add_channel("INDEX", data=np.arange(10), cast_dtype=np.uint8)
    ch2 = lf.add_channel("X", data=np.arange(-10, 20).astype(np.int16))
    ch3 = lf.add_channel(
        "Y", data=np.random.randint(-10, 11, size=(10, 20)), cast_dtype=np.int32
    )
    ch4 = lf.add_channel("Z", data=np.arange(10), cast_dtype=np.int8)
    lf.add_frame(
        "MAIN",
        channels=(ch1, ch2, ch3, ch4),
        index_type=enums.FrameIndexType.NON_STANDARD,
    )

    return df


def test_sint_data(caplog: pytest.LogCaptureFixture) -> None:

    with caplog.at_level(logging.WARNING, logger="dliswriter"):
        df = _prepare_file_with_sint_data()
        df.generate_logical_records(None)
        assert (
            "Data type of channel 'X' is int16; some DLIS viewers cannot interpret signed integers"
            in caplog.text
        )
        assert (
            "Data type of channel 'Y' is int32; some DLIS viewers cannot interpret signed integers"
            in caplog.text
        )
        assert (
            "Data type of channel 'Z' is int8; some DLIS viewers cannot interpret signed integers"
            in caplog.text
        )


@high_compatibility_mode_decorator
def test_sint_data_high_compat_mode() -> None:
    df = _prepare_file_with_sint_data()

    with pytest.raises(
        RuntimeError,
        match="Data type of channel 'X' is int16; some DLIS viewers cannot interpret signed integers.*",
    ):
        df.generate_logical_records(None)



import numpy as np
import h5py    # type: ignore  # untyped library
from typing import Union, Optional, Any, Generator
import logging
from abc import ABC

from dliswriter.utils.internal.converters import ReprCodeConverter
from dliswriter.utils.internal.types import data_form_type, data_source_type, file_name_type, numpy_dtype_type


logger = logging.getLogger(__name__)


class SourceDataWrapper(ABC):
    """Keep reference to source data. Produce chunks of input data as asked, in the form of a structured numpy array."""

    def __init__(self, data_source: data_source_type, mapping: dict[str, str],
                 known_dtypes: Optional[dict[str, numpy_dtype_type]] = None, from_idx: int = 0,
                 to_idx: Optional[int] = None) -> None:
        """Initialise a SourceDataWrapper.

        Args:
            data_source     :   Original data object.
            mapping         :   Mapping of data type names on the names of data in the data source (e.g. on the paths to
                                particular HDF5 datasets).
            known_dtypes    :   Mapping of data type names on data types (if any are known). Does not have to contain
                                all dtypes. Can also be completely omitted. Missing data types are determined from
                                the data.
            from_idx        :   Index from which data should be loaded (or number of initial rows to ignore).
            to_idx          :   Index up to which data should be loaded.

            Note:
                All data sets from 'mapping' should be found in the 'data_source'. On the other hand, 'data_source'
                can contain more data sets than mentioned in 'mapping' and the order does not have to match.
                Only the data sets found in 'mapping', with that exact order, will be included in the produced
                data chunks, passed to the DLIS file writer loop.
                All datasets (of the ones mentioned in 'mapping') are required to have the same length.
        """

        self._data_source = data_source
        self._mapping = mapping

        # numpy dtype object which will be used for constructing data chunks (see 'load_chunk')
        self._dtype = self.determine_dtypes(self._data_source, self._mapping, known_dtypes=known_dtypes)

        # total number of rows - guessed from the first dataset
        total_n_rows = self._data_source[next(iter(mapping.values()))].shape[0]

        self._from_idx = from_idx
        self._to_idx = to_idx if to_idx is not None else total_n_rows
        self._n_rows = self._to_idx - self._from_idx  # number of rows to be loaded

        if self._from_idx >= total_n_rows:
            raise ValueError(f"Starting index {self._from_idx} too large for total n. rows {total_n_rows}")
        if self._n_rows < 1:
            raise ValueError(f"Starting index {self._from_idx} and end index {self._to_idx} do not yield a positive "
                             f"number of rows to be loaded")

    @property
    def n_rows(self) -> int:
        """Total number of data rows."""

        return self._n_rows

    @property
    def data_source(self) -> data_source_type:
        """Source data object."""

        return self._data_source

    @property
    def dtype(self) -> np.dtype:
        """Data type of the structured numpy arrays, in the form of which the input data chunks are loaded."""

        return self._dtype

    @staticmethod
    def determine_dtypes(data_object: data_source_type, mapping: dict[str, str],
                         known_dtypes: Optional[dict[str, numpy_dtype_type]] = None) -> np.dtype:
        """Determine the structured numpy dtype for the provided data, with given data names mapping.

        Args:
            data_object     :   Original data object.
            mapping         :   Mapping of data type names on the names of data in the data source (e.g. on the paths to
                                particular HDF5 datasets).
            known_dtypes    :   Mapping of data type names on data types (if any are known). Does not have to contain
                                all dtypes. Can also be completely omitted. Missing data types are determined from
                                the data.

        Returns:
            A numpy.dtype object, specifying the names, dtypes, and (if not 1D) shapes of the data sets to be contained
            in the loaded input data chunks.
        """

        dtypes = []  # list of data type tuples, later transformed to a np.dtype
        known_dtypes = known_dtypes or {}

        for dtype_name, dataset_name in mapping.items():
            dt: Union[tuple[str, Any], tuple[str, Any, int]]

            try:
                dset = data_object[dataset_name]
            except (ValueError, KeyError):
                raise ValueError(f"No dataset '{dataset_name}' found in the source data")
            dset_row0 = dset[0:1]  # get in form of a 1-element (or 1-row) array, not a single number
            # for h5 data, the above retrieves the first row of the current data set

            # determine the numpy number dtype
            number_type = known_dtypes.get(dtype_name, dset_row0.dtype)
            ReprCodeConverter.validate_numpy_dtype(number_type)

            # determine the dtype of the data set (2- or 3-tuple)
            dt = (dtype_name, number_type)
            if dset_row0.ndim > 1:
                if dset_row0.ndim > 2:
                    raise RuntimeError("Data sets with more than 2 dimensions are not supported")
                dt = (*dt, dset_row0.shape[-1])  # 3-tuple if the data set has multiple samples per row - add the width
            dtypes.append(dt)

        return np.dtype(dtypes)

    def __getitem__(self, item: str) -> np.ndarray:
        """Retrieve a dataset of the given name from the dataset.

        The name should be the one used for the dtype name, not the data set name/location
        (i.e. taken from the keys, not values, of the 'mapping' dictionary specified at init).

        Returns:
            A np.ndarray with the data.
        """

        try:
            data = self._data_source[self._mapping[item]]
        except (ValueError, KeyError):
            raise ValueError(f"No dataset '{item}' found in the source data")
        return data[self._from_idx:self._to_idx]

    def load_chunk(self, start: int, stop: Union[int, None]) -> np.ndarray:
        """Copy a chunk of the source data into a structured numpy array of the pre-determined dtype.

        The returned chunk is a horizontal slice of the source data, possibly reshaped to match the desired format
        (i.e. the order and names of data sets). All data sets specified in the initial mapping are included.
        The range of the included rows of data is determined by the 'start' and 'stop' arguments.

        Args:
            start   :   Start index.
            stop    :   Stop index. If None, all data from start index till the end will be loaded.

        Returns:
            A structured numpy array, containing the required chunks of all the relevant data sets from the source data.
        """

        if start < 0:
            raise ValueError("Start row cannot be negative")

        if start > self._n_rows:
            raise ValueError(f"Cannot load chunk from row {start} because the data source has only {self._n_rows} rows")

        if stop is None:
            stop = self._n_rows

        if stop > self._n_rows:
            raise ValueError(f"Cannot load chunk up to row {stop} because the data source has only {self._n_rows} rows")
        if stop < start:
            raise ValueError(f"Stop row cannot be smaller than start row; got {stop} and {start}")

        idx = slice(self._from_idx + start, self._from_idx + stop)
        n_rows = stop - start

        chunk = np.zeros(n_rows, dtype=self._dtype)
        for key, loc in self._mapping.items():
            chunk[key] = self._data_source[loc][idx]

        return chunk

    def make_chunked_generator(self, chunk_rows: Union[int, None]) -> Generator:
        """Define a generator yielding consecutive chunks of input data with the specified size.

        Args:
            chunk_rows  :   Maximal number of rows per chunk (the last chunk might be smaller, depending on the total
                            size of the data). If None, the entire data is loaded as a single chunk.

        Yields:
            Structured numpy.ndarray objects with the consecutive chunks of the source data.
        """

        if chunk_rows is None:
            chunk_rows = self._n_rows
            n_full_chunks = 1
            remainder_rows = 0
            logger.debug(f"Data will be loaded in a single chunk of {self._n_rows}")
        else:
            n_full_chunks, remainder_rows = divmod(self._n_rows, chunk_rows)
            if n_full_chunks:
                rem = f" plus a last, smaller chunk of {remainder_rows} rows" if remainder_rows else ""
                logger.debug(f"Data will be loaded in {n_full_chunks} chunk(s) of {chunk_rows} rows" + rem)
            else:
                logger.debug(f"Provided chunk size ({chunk_rows}) is larger than the total size of the data "
                             f"({self._n_rows}); data will be loaded in a single chunk of {remainder_rows}")

        total_chunks = n_full_chunks + int(bool(remainder_rows))

        for i in range(n_full_chunks):
            logger.debug(f"Loading chunk {i+1}/{total_chunks} ({chunk_rows} rows)")
            yield from self.load_chunk(i * chunk_rows, (i + 1) * chunk_rows)

        if remainder_rows:
            logger.debug(f"Loading chunk {total_chunks}/{total_chunks} ({remainder_rows} rows)")
            yield from self.load_chunk(n_full_chunks * chunk_rows, None)

    @classmethod
    def make_wrapper(cls, source: data_form_type, mapping: Optional[dict] = None,
                     **kwargs: Any) -> Union["DictDataWrapper", "NumpyDataWrapper", "HDF5DataWrapper"]:
        """Create an instance of one of the SourceDataWrapper subclasses based on the provided data.

        Args:
            source  :   Original data object.
            mapping :   Mapping of data type names on the names of data in the data source (e.g. on the paths to
                        particular HDF5 datasets).
            kwargs:     Additional keyword arguments accepted by the SourceDataWrapper subclasses' constructors.
        """

        if isinstance(source, dict):
            return DictDataWrapper(source, mapping, **kwargs)

        if isinstance(source, np.ndarray):
            return NumpyDataWrapper(source, mapping, **kwargs)

        try:
            source_str = str(source)
        except (TypeError, ValueError):
            raise TypeError(f"Expected a path-like; got {type(source)}: {source}")
        if source_str.split('.')[-1].lower() not in ('h5', 'hdf5'):
            raise ValueError(f"Expected a path to an HDF5 file; got {source_str}")
        if mapping is None:
            raise ValueError("Mapping must be provided to create a HDF5DataWrapper")
        return HDF5DataWrapper(source, mapping, **kwargs)


class HDF5DataWrapper(SourceDataWrapper):
    """Wrap source data provided in the form of a HDF5 file."""

    _data_source: h5py.File

    def __init__(self, data_file_name: file_name_type, mapping: dict,
                 known_dtypes: Optional[dict[str, numpy_dtype_type]] = None, from_idx: int = 0,
                 to_idx: Optional[int] = None) -> None:
        """Initialise HDF5DataWrapper.

        Args:
            data_file_name  :   Name of/path to the HDF5 file containing the source data.
            mapping         :   Mapping of the names of data sets (data types) to be included on the corresponding
                                data set paths in the file.
            known_dtypes    :   Mapping of data type names on data types (if any are known). Does not have to contain
                                all dtypes. Can also be completely omitted. Missing data types are determined from
                                the data.
            from_idx        :   Index from which data should be loaded (or number of initial rows to ignore).
            to_idx          :   Index up to which data should be loaded.
        """

        # open the file
        h5_data = h5py.File(data_file_name, 'r')

        # add a forward slash at the beginning of each value in the mapping dict - if missing
        mapping = {k: (f'/{v}' if not v.startswith('/') else v) for k, v in mapping.items()}

        super().__init__(h5_data, mapping, known_dtypes=known_dtypes, from_idx=from_idx, to_idx=to_idx)

    def close(self) -> None:
        """Close the HDF5 file (if open)."""

        if hasattr(self, '_data_source'):  # object might be partially initialised
            try:
                self._data_source.close()
            except TypeError as exc:
                logger.error(f"Error closing the source data file: {exc}")
            else:
                logger.debug("Source data file closed")

    def __del__(self) -> None:
        """Close the HDF5 file when deleting the object (if still open at this point)."""

        self.close()


class NumpyDataWrapper(SourceDataWrapper):
    """Wrap source data provided in the form of a structured numpy array."""

    _data_source: np.ndarray

    def __init__(self, arr: np.ndarray, mapping: Optional[dict] = None,
                 known_dtypes: Optional[dict[str, numpy_dtype_type]] = None, from_idx: int = 0,
                 to_idx: Optional[int] = None) -> None:
        """Initialise NumpyDataWrapper.

        Args:
            arr             :   Source data - structured numpy array.
            mapping         :   Mapping of target data type names on the data type names found in the source array.
                                Optional; if not provided, it is assumed that the target data types (data sets)
                                are the same as the source ones.
            known_dtypes    :   Mapping of data type names on data types (if any are known). Does not have to contain
                                all dtypes. Can also be completely omitted. Missing data types are determined from
                                the data.
            from_idx        :   Index from which data should be loaded (or number of initial rows to ignore).
            to_idx          :   Index up to which data should be loaded.

        """

        self._check_source_arr(arr)

        if not mapping:
            # default mapping: 1 to 1 for all existing data type names
            mapping = {k: k for k in arr.dtype.names}

        super().__init__(arr, mapping, known_dtypes=known_dtypes, from_idx=from_idx, to_idx=to_idx)

    def load_chunk(self, start: int, stop: Union[int, None]) -> np.ndarray:
        """Load a chunk of the input data.

        If the target data type (data sets names and dtypes) is the same as the one in the source array,
        take a slice of the source array directly.
        Otherwise, use the 'load_chunk' from the superclass to copy the relevant data sets into a new structured array.

        Args:
            start   :   Start index.
            stop    :   Stop index. If None, all data from start index till the end will be loaded.

        Returns:
            A structured numpy array, containing the required chunks of all the relevant data sets from the source data.
        """

        if self._dtype == self._data_source.dtype:
            return self._data_source[start:stop]

        return super().load_chunk(start, stop)

    @staticmethod
    def _check_source_arr(arr: np.ndarray) -> None:
        """Check that the input array is a structured, 1D numpy array."""

        if not isinstance(arr, np.ndarray):
            raise TypeError(f"Expected a numpy.ndarray; got {type(arr)}")

        if arr.ndim > 1:
            raise ValueError("Source array must be 1-dimensional")

        if not arr.dtype.names:
            raise ValueError("Input must be a structured numpy array")


class DictDataWrapper(SourceDataWrapper):
    """Wrap source data provided in the form of a dictionary of numpy arrays."""

    def __init__(self, data_dict: dict[str, np.ndarray], mapping: Optional[dict] = None,
                 known_dtypes: Optional[dict[str, numpy_dtype_type]] = None, from_idx: int = 0,
                 to_idx: Optional[int] = None) -> None:
        """Initialise DictDataWrapper.

        Args:
            data_dict       :   Source data - dict of numpy arrays.
            mapping         :   Mapping of target data type names on the keys found in the data dictionary.
                                Optional; if not provided, it is assumed that all items of the data dict should be
                                included in the target structured arrays.
            known_dtypes    :   Mapping of data type names on data types (if any are known). Does not have to contain
                                all dtypes. Can also be completely omitted. Missing data types are determined from
                                the data.
            from_idx        :   Index from which data should be loaded (or number of initial rows to ignore).
            to_idx          :   Index up to which data should be loaded.

        """

        self._check_source_dict(data_dict)

        if not mapping:
            # default mapping: 1 to 1 for all keys of the data dict
            mapping = {k: k for k in data_dict.keys()}

        super().__init__(data_dict, mapping, known_dtypes=known_dtypes, from_idx=from_idx, to_idx=to_idx)

    @staticmethod
    def _check_source_dict(data_dict: dict[str, np.ndarray]) -> None:
        """Check that all values of the source dictionary are numpy arrays."""

        if not isinstance(data_dict, dict):
            raise TypeError(f"Expected a dictionary, got {type(data_dict)}: {data_dict}")

        if not all(isinstance(k, str) for k in data_dict):
            raise TypeError(f"Source dictionary keys must be strings; got {', '.join(str(type(k)) for k in data_dict)}")

        if not all(isinstance(v, np.ndarray) for v in data_dict.values()):
            raise TypeError(f"Dict values must be numpy arrays; "
                            f"got {', '.join(str(type(v)) for v in data_dict.values())}")



import logging
from typing import Any

from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem
from dliswriter.logical_record.eflr_types.equipment import EquipmentSet
from dliswriter.logical_record.eflr_types.channel import ChannelSet
from dliswriter.logical_record.eflr_types.parameter import ParameterSet
from dliswriter.utils.internal.internal_enums import EFLRType
from dliswriter.logical_record.core.attribute import EFLRAttribute, StatusAttribute, TextAttribute


logger = logging.getLogger(__name__)


class ToolItem(EFLRItem):
    """Model an object being part of Tool EFLR."""

    parent: "ToolSet"

    def __init__(self, name: str, parent: "ToolSet", **kwargs: Any) -> None:
        """Initialise ToolItem.

        Args:
            name        :   Name of the ToolItem.
            parent      :   Parent ToolSet of this ToolItem.
            **kwargs    :   Values of to be set as characteristics of the ToolItem Attributes.
        """

        self.description = TextAttribute('description')
        self.trademark_name = TextAttribute('trademark_name')
        self.generic_name = TextAttribute('generic_name')
        self.parts = EFLRAttribute('parts', object_class=EquipmentSet, multivalued=True)
        self.status = StatusAttribute('status')
        self.channels = EFLRAttribute('channels', object_class=ChannelSet, multivalued=True)
        self.parameters = EFLRAttribute('parameters', object_class=ParameterSet, multivalued=True)

        super().__init__(name, parent=parent, **kwargs)


class ToolSet(EFLRSet):
    """Model Tool EFLR."""

    set_type = 'TOOL'
    logical_record_type = EFLRType.STATIC
    item_type = ToolItem


ToolItem.parent_eflr_class = ToolSet



from dliswriter.logical_record.core.attribute import EFLRAttribute
from dliswriter.logical_record.eflr_types import (ChannelItem, ParameterItem, CalibrationMeasurementItem,
                                                  CalibrationCoefficientItem, CalibrationItem, CalibrationSet)


def _check_list(objects: EFLRAttribute, names: tuple[str, ...], object_class: type) -> None:
    """Check that the objects defined for a given EFLR attribute match the expected names and type."""

    objects_list = objects.value

    assert isinstance(objects_list, list)
    assert len(objects_list) == len(names)
    for i, name in enumerate(names):
        assert isinstance(objects_list[i], object_class)
        assert objects_list[i].name == name


def test_calibration_creation(channel1: ChannelItem, channel2: ChannelItem, channel3: ChannelItem,
                              ccoef1: CalibrationCoefficientItem, cmeasure1: CalibrationMeasurementItem,
                              param1: ParameterItem, param2: ParameterItem, param3: ParameterItem) -> None:
    """Check that a CalibrationObject is correctly set up from config info."""

    c = CalibrationItem(
        "CALIB-MAIN",
        calibrated_channels=(channel1, channel2),
        uncalibrated_channels=(channel3,),
        coefficients=(ccoef1,),
        measurements=(cmeasure1,),
        parameters=(param1, param2, param3),
        parent=CalibrationSet()
    )

    assert c.name == "CALIB-MAIN"

    _check_list(c.calibrated_channels, ("Channel 1", "Channel 2"), ChannelItem)
    _check_list(c.uncalibrated_channels, ("Channel 3",), ChannelItem)
    _check_list(c.coefficients, ("COEF-1",), CalibrationCoefficientItem)
    _check_list(c.measurements, ("CMEASURE-1",), CalibrationMeasurementItem)
    _check_list(c.parameters, ("Param-1", "Param-2", "Param-3"), ParameterItem)



import logging
from typing import Any

from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem
from dliswriter.utils.internal.internal_enums import EFLRType
from dliswriter.logical_record.core.attribute import NumericAttribute, IdentAttribute
from dliswriter.utils.internal.value_checkers import validate_string


logger = logging.getLogger(__name__)


class CalibrationCoefficientItem(EFLRItem):
    """Model an object being part of CalibrationCoefficient EFLR."""

    parent: "CalibrationCoefficientSet"

    def __init__(self, name: str, parent: "CalibrationCoefficientSet", **kwargs: Any) -> None:
        """Initialise CalibrationCoefficientItem.

        Args:
            name        :   Name of the CalibrationCoefficientItem.
            parent      :   Parent CalibrationCoefficientSet of this CalibrationCoefficientItem.
            **kwargs    :   Values of to be set as characteristics of the CalibrationCoefficientItem Attributes.
        """

        self.label = IdentAttribute('label', converter=validate_string)
        self.coefficients = NumericAttribute('coefficients', multivalued=True)
        self.references = NumericAttribute('references', multivalued=True)
        self.plus_tolerances = NumericAttribute('plus_tolerances', multivalued=True)
        self.minus_tolerances = NumericAttribute('minus_tolerances', multivalued=True)

        super().__init__(name, parent=parent, **kwargs)

    def _run_checks_and_set_defaults(self) -> None:
        """Check that the number of coefficients, references, and tolerances is equal."""

        value_counts = self.count_attributes(
            self.coefficients, self.references, self.plus_tolerances, self.minus_tolerances)

        if len(set(value_counts.values())) > 1:
            raise RuntimeError(f"Number of values all numeric attributes of Calibration Coefficient should be equal; "
                               f"got {', '.join('{} for {}'.format(v, k) for k, v in value_counts.items())}")


class CalibrationCoefficientSet(EFLRSet):
    """Model CalibrationCoefficient EFLR."""

    set_type = 'CALIBRATION-COEFFICIENT'
    logical_record_type = EFLRType.STATIC
    item_type = CalibrationCoefficientItem


CalibrationCoefficientItem.parent_eflr_class = CalibrationCoefficientSet



from dliswriter.logical_record.eflr_types import SpliceSet, SpliceItem, ZoneItem, ChannelItem


def test_splice_creation(zone1: ZoneItem, zone3: ZoneItem, chan: ChannelItem, channel2: ChannelItem,
                         channel1: ChannelItem, channel3: ChannelItem) -> None:
    """Test creating SpliceItem."""

    splice = SpliceItem(
        "splc1",
        zones=(zone1, zone3),
        input_channels=(channel2, channel1, chan),
        output_channel=channel3,
        parent=SpliceSet()
    )

    assert splice.name == "splc1"

    for i, n in enumerate(("Zone-1", "Zone-3")):
        assert splice.zones.value[i].name == n

    for i, n in enumerate(("Channel 2", "Channel 1", "some_channel")):
        assert splice.input_channels.value[i].name == n

    assert splice.output_channel.value.name == 'Channel 3'

    assert isinstance(splice.parent, SpliceSet)
    assert splice.parent.set_name is None



import numpy as np
import pytest
from typing import Union, Any

from dliswriter.utils.source_data_wrappers import NumpyDataWrapper, SourceDataWrapper


@pytest.fixture
def dtype() -> np.dtype:
    data_types = [
        ('depth', float),
        ('rpm', np.int32),
        ('amplitude', np.float32, 128),
        ('radius', np.int32, 20)
    ]
    return np.dtype(data_types)


def test_basic_properties(data: np.ndarray) -> None:
    w = NumpyDataWrapper(data)
    assert w.data_source is data
    assert w.n_rows == 50
    assert isinstance(w.dtype, np.dtype)
    assert isinstance(w.dtype.names, tuple)
    assert len(w.dtype.names) == 4


@pytest.fixture
def data(dtype: np.dtype) -> np.ndarray:
    n = 50
    arr = np.empty(n, dtype=dtype)
    arr['depth'] = np.arange(n) * 0.01
    arr['rpm'] = np.random.randint(size=n, low=5, high=10, dtype=np.int32)
    arr['amplitude'] = np.random.rand(n, 128)
    arr['radius'] = (10 * np.random.rand(n, 20)).astype(np.int32)

    return arr


def test_creation(data: np.ndarray) -> None:
    w = NumpyDataWrapper(data)

    assert w.dtype == data.dtype
    assert w.dtype.names == ('depth', 'rpm', 'amplitude', 'radius')
    assert w.dtype[0] == np.float64
    assert w.dtype[1] == np.int32
    assert w.dtype[2] == (np.float32, (128,))
    assert w.dtype[3] == (np.int32, (20,))


def test_creation_with_mapping(data: np.ndarray) -> None:
    w = NumpyDataWrapper(data, mapping={'DPTH': 'depth', 'RPM': 'rpm', 'AMP': 'amplitude', 'RAD': 'radius'})

    assert w.dtype.names == ('DPTH', 'RPM', 'AMP', 'RAD')
    assert w.dtype[0] == np.float64
    assert w.dtype[1] == np.int32
    assert w.dtype[2] == (np.float32, (128,))
    assert w.dtype[3] == (np.int32, (20,))


def test_creation_with_mapping_different_order(data: np.ndarray) -> None:
    w = NumpyDataWrapper(data, mapping={'RAD': 'radius', 'AMP': 'amplitude', 'RPM': 'rpm', 'DPTH': 'depth'})

    assert w.dtype.names == ('RAD', 'AMP', 'RPM', 'DPTH')
    assert w.dtype[3] == np.float64
    assert w.dtype[2] == np.int32
    assert w.dtype[1] == (np.float32, (128,))
    assert w.dtype[0] == (np.int32, (20,))


def test_creation_with_mapping_omitting_dsets(data: np.ndarray) -> None:
    w = NumpyDataWrapper(data, mapping={'AMP': 'amplitude', 'DPTH': 'depth'})

    assert w.dtype.names == ('AMP', 'DPTH')
    assert w.dtype[1] == np.float64
    assert w.dtype[0] == (np.float32, (128,))


@pytest.mark.parametrize(('known_dtypes', 'dtype_check'), (
        ({'depth': np.float32, 'rpm': np.float64}, (np.float32, np.float64, np.float32, np.int32)),
        ({'rpm': np.int32}, (np.float64, np.int32, np.float32, np.int32)),
        ({'radius': np.float32, 'amplitude': np.float64}, (np.float64, np.int32, np.float64, np.float32))
))
def test_creation_with_known_dtypes(data: np.ndarray, known_dtypes: dict, dtype_check: tuple) -> None:
    w = NumpyDataWrapper(data, known_dtypes=known_dtypes)

    assert w.dtype[0] == dtype_check[0]
    assert w.dtype[1] == dtype_check[1]
    assert w.dtype[2] == (dtype_check[2], (128,))
    assert w.dtype[3] == (dtype_check[3], (20,))


def test_creation_with_known_dtypes_and_mapping(data: np.ndarray) -> None:
    w = NumpyDataWrapper(
        data,
        mapping={'RAD': 'radius', 'AMP': 'amplitude', 'RPM': 'rpm'},
        known_dtypes={'AMP': np.int32, 'RPM': np.float32}
    )

    assert w.dtype.names == ('RAD', 'AMP', 'RPM')
    assert w.dtype[0] == (np.int32, (20,))  # radius
    assert w.dtype[1] == (np.int32, (128,))  # amplitude
    assert w.dtype[2] == np.float32  # rpm


@pytest.mark.parametrize('dat', (
        '/path/to/file.npz',
        {'depth': np.arange(50), 'amplitude': np.random.rand(50, 128)},
        [1, 2, 3, 4, 5]
))
def test_type_error_if_not_numpy(dat: Any) -> None:
    with pytest.raises(TypeError, match="Expected a numpy.ndarray.*"):
        NumpyDataWrapper(dat)


@pytest.mark.parametrize('dat', (
        np.arange(12),
        np.random.rand(50),
        np.random.rand(150)
))
def test_value_error_if_not_structured_numpy(dat: np.ndarray) -> None:
    with pytest.raises(ValueError, match="Input must be a structured numpy array"):
        NumpyDataWrapper(dat)


@pytest.mark.parametrize('dat', (
        np.random.rand(50, 150),
        np.random.rand(150, 50)
))
def test_value_error_if_array_not_1d(dat: np.ndarray) -> None:
    with pytest.raises(ValueError, match="Source array must be 1-dimensional"):
        NumpyDataWrapper(dat)


@pytest.mark.parametrize('shape', ((10, 20), (1, 50), (12, 17, 19)))
def test_value_error_if_structured_array_not_1d(dtype: np.dtype, shape: tuple) -> None:
    with pytest.raises(ValueError, match="Source array must be 1-dimensional"):
        NumpyDataWrapper(np.ones(shape, dtype=dtype))


def test_creation_from_superclass(data: np.ndarray) -> None:
    w = SourceDataWrapper.make_wrapper(data)
    assert isinstance(w, NumpyDataWrapper)


@pytest.mark.parametrize(('from_idx', 'to_idx', 'n_rows'), ((0, 20, 20), (32, None, 18), (12, 25, 13)))
def test_creation_with_from_and_to_idx(data: np.ndarray, from_idx: int, to_idx: Union[int, None], n_rows: int) -> None:
    w = NumpyDataWrapper(data, from_idx=from_idx, to_idx=to_idx)
    assert w.n_rows == n_rows


@pytest.mark.parametrize(('start', 'stop'), ((0, 10), (1, 2), (3, 3), (3, 17)))
def test_load_chunk(data: np.ndarray, start: int, stop: int) -> None:
    w = NumpyDataWrapper(data)
    chunk = w.load_chunk(start, stop)

    assert chunk.size == stop - start
    assert chunk.dtype == w.dtype
    assert (chunk['depth'] == data['depth'][start:stop]).all()
    assert (chunk['rpm'] == data['rpm'][start:stop]).all()
    assert (chunk['amplitude'] == data['amplitude'][start:stop]).all()


@pytest.mark.parametrize(('start', 'stop'), ((30, 45), (11, 12)))
def test_load_chunk_alternative_mapping(data: np.ndarray, start: int, stop: int) -> None:
    w = NumpyDataWrapper(data, mapping={'RPM': 'rpm', 'MD': 'depth'})  # no amplitude, switched order
    chunk = w.load_chunk(start, stop)

    assert chunk.size == stop - start
    assert chunk.dtype.names == ('RPM', 'MD')
    assert chunk.dtype == w.dtype

    assert (chunk['RPM'] == data['rpm'][start:stop]).all()
    assert (chunk['MD'] == data['depth'][start:stop]).all()


@pytest.mark.parametrize(("from_idx", "to_idx"), ((0, 30), (40, None)))
def test_getitem_default_mapping(data: np.ndarray, from_idx: int, to_idx: Union[int, None]) -> None:
    w = NumpyDataWrapper(data, from_idx=from_idx, to_idx=to_idx)

    for key in ('depth', 'amplitude', 'rpm', 'radius'):
        assert isinstance(w[key], np.ndarray)

    assert (w['depth'] == data['depth'][from_idx:to_idx]).all()
    assert (w['amplitude'] == data['amplitude'][from_idx:to_idx]).all()
    assert (w['rpm'] == data['rpm'][from_idx:to_idx]).all()
    assert (w['radius'] == data['radius'][from_idx:to_idx]).all()


@pytest.mark.parametrize(("from_idx", "to_idx"), ((1, None), (11, 12)))
def test_getitem_alternative_mapping(data: np.ndarray, from_idx: int, to_idx: Union[int, None]) -> None:
    w = NumpyDataWrapper(data, from_idx=from_idx, to_idx=to_idx, mapping={'AMP': 'amplitude', 'RAD': 'radius'})

    for key in ('AMP', 'RAD'):
        assert isinstance(w[key], np.ndarray)

    assert (w['AMP'] == data['amplitude'][from_idx:to_idx]).all()
    assert (w['RAD'] == data['radius'][from_idx:to_idx]).all()

    for key in ('depth', 'rpm', 'amplitude', 'radius'):
        with pytest.raises(ValueError, match=f"No dataset '{key}' found in the source data"):
            w[key]



from dlisio import dlis    # type: ignore  # untyped library

from tests.common import check_list_of_objects


def test_splices(short_dlis: dlis.file.LogicalFile) -> None:
    """Check that the number of Splice objects in the DLIS file matches the expected one."""

    splices = short_dlis.splices
    assert len(splices) == 1


def test_splice_params(short_dlis: dlis.file.LogicalFile) -> None:
    """Test attributes of the Splice object in the new DLIS file."""

    splice = short_dlis.splices[0]

    assert splice.name == "splc1"
    assert splice.origin == 42

    check_list_of_objects(splice.zones, ("Zone-1", "Zone-2"))
    check_list_of_objects(splice.input_channels, ("Channel 1", "Channel 2"))

    assert splice.output_channel.name == 'amplitude'



"""Create an exemplary DLIS file with synthetic data and all kinds of DLIS objects supported in this library."""

from datetime import datetime, timedelta
import numpy as np
import logging

from dliswriter import AttrSetup, DLISFile, enums
from dliswriter.file.file import LogicalFile

from utils import install_colored_logger


# colored logs output
install_colored_logger(logging.getLogger("dliswriter"))


# create DLISFile instance; optionally, pass custom parameters for file header and storage unit label
df = DLISFile()
lf: LogicalFile = df.add_logical_file()

# add origin
origin = lf.add_origin("DEFAULT ORIGIN", company="XXX", order_number="352")
origin.creation_time.value = datetime(
    year=2023, month=12, day=6, hour=12, minute=30, second=5
)

# multiple origins can be added;
# the origin_reference of the first origin will be used as origin_reference in all objects automatically
# but you can also choose to pass other origins' origin_reference as the reference
# to indicate that a given object belongs to one of the other origins
origin2 = lf.add_origin("ADDITIONAL ORIGIN", well_name="XYZ", company="ABC")
origin3 = lf.add_origin(
    "ANOTHER ORIGIN", well_name="XYZ", company="another company", origin_reference=35
)


# define axes - metadata objects for channels
ax1 = lf.add_axis(
    "AXIS1", coordinates=["40 23' 42.8676'' N", "27 47' 32.8956'' E"], axis_id="AXIS 1"
)
ax1.spacing.value = 0.2
ax1.spacing.units = enums.Unit.METER

ax2 = lf.add_axis(
    "AXIS2",
    spacing=5,
    coordinates=[1, 2, 3.5],
    origin_reference=origin2.origin_reference,  # mark ax2 as belonging to origin2
)

ax3 = lf.add_axis("AXIS3", spacing=0.1, coordinates=[0])


# define long_names - descriptions for channels, parameters, and computations
long_name1 = lf.add_long_name("LONG-NAME1", quantity="23", standard_symbol="ABC")
long_name2 = lf.add_long_name(
    "LONG-NAME-2", entity_part="X", source_part_number=["121.111"]
)
long_name3 = lf.add_long_name(
    "ANOTHER LONG NAME", conditions=["At Standard Temperature"]
)


# define frame 1: depth-based with 4 channels, 100 rows each
n_rows_depth = 100
ch1 = lf.add_channel(
    "DEPTH",
    data=np.arange(n_rows_depth) / 10
    - 3,  # index channel - always scalar, i.e. 1D data
    units=enums.Unit.METER,
)
ch2 = lf.add_channel(
    "RPM", data=(np.arange(n_rows_depth) % 10).astype(np.int32) - 2, axis=ax3  # 1D data
)
ch3 = lf.add_channel(
    "AMPLITUDE",
    data=np.random.rand(n_rows_depth, 5),  # 2D data - 5 columns
    cast_dtype=np.float32,
    long_name=long_name3,
)
ch4 = lf.add_channel(
    "COMPUTED_CHANNEL",
    data=np.random.randint(0, 100, dtype=np.uint8, size=n_rows_depth),
    long_name=long_name1,
)
main_frame = lf.add_frame(
    "MAIN FRAME",
    channels=(ch1, ch2, ch3, ch4),
    index_type=enums.FrameIndexType.BOREHOLE_DEPTH,
)


# define frame 2: time-based with 2 channels, 200 rows each
n_rows_time = 200
ch5 = lf.add_channel(
    # index channel for frame 2
    "TIME",
    data=np.arange(n_rows_time),
    cast_dtype=np.uint32,
    units=enums.Unit.SECOND,
    axis=ax3,
)
ch6 = lf.add_channel(
    "TEMPERATURE",
    data=np.random.randint(-10, 30, size=n_rows_time, dtype=np.int8),
    cast_dtype=np.int16,
    units=enums.Unit.DEGREE_CELSIUS,
)
second_frame = lf.add_frame(
    "TIME FRAME", channels=(ch5, ch6), index_type=enums.FrameIndexType.NON_STANDARD
)


# zones
zone1 = lf.add_zone(
    "DEPTH-ZONE", domain=enums.ZoneDomain.BOREHOLE_DEPTH, minimum=2, maximum=4.5
)
dt = datetime.now()
zone2 = lf.add_zone(
    "TIME-ZONE",
    domain=enums.ZoneDomain.TIME,
    minimum=dt - timedelta(hours=3),
    maximum=dt - timedelta(minutes=30),
    origin_reference=origin3.origin_reference,  # associated with origin 3
)
zone3 = lf.add_zone(
    "VDEPTH-ZONE",
    domain=enums.ZoneDomain.VERTICAL_DEPTH,
    minimum=10,
    maximum=20,
    origin_reference=origin2.origin_reference,  # associated with origin 2
)


# splices - using zones & channels
splice1 = lf.add_splice(
    "SPLICE1", input_channels=(ch1,), output_channel=ch4, zones=(zone1,)
)
splice2 = lf.add_splice(
    "SPLICE2", input_channels=(ch5, ch2), output_channel=ch6, zones=(zone2, zone3)
)


# parameters - using zones, axes, and long name
parameter1 = lf.add_parameter(
    "PARAM1",
    long_name="Parameter nr 1",  # long_name can be str or a LongName object
    axis=ax1,
    values={
        "value": [1],
        "units": enums.Unit.INCH,
    },  # specifying value and units of the value in one line
)
parameter2 = lf.add_parameter(
    "PARAM2",
    zones=(zone2,),
    long_name=long_name2,
    values=[
        ["val1", "val2", "val3"]
    ],  # specifying only values; units can be added by: parameter2.values.units = ...
    dimension=[3],
)


# equipment
# note how complex attribute values can be passed as a dict or an AttrSetup object; the effect exactly is the same
equipment1 = lf.add_equipment(
    "EQ1",
    status=1,
    eq_type=enums.EquipmentType.TOOL,
    serial_number="1239-12312",
    weight={
        "value": 123.2,
        "units": enums.Unit.KILOGRAM,
    },  # specifying value and units in one line - dict version
    length=AttrSetup(
        2, enums.Unit.METER
    ),  # specifying value and units in one line - AttrSetup version
)

# 'value' and 'units' can also be added later to the created object.
equipment2 = lf.add_equipment(
    "EQ2", location=enums.EquipmentLocation.WELL, trademark_name="Some trademark TM"
)
equipment2.hole_size.value = 23.5
equipment2.hole_size.units = enums.Unit.INCH

equipment3 = lf.add_equipment("EQ3")
equipment3.status.value = 0


# tool - using equipment, channels, and parameters
tool1 = lf.add_tool(
    "TOOL1",
    status=1,
    parts=(equipment1, equipment2),
    channels=(ch5, ch6),
    description="...",
)
tool2 = lf.add_tool(
    "TOOL2",
    parameters=(parameter1, parameter2),
    channels=(ch1, ch2, ch3),
    parts=(equipment1,),
)


# computation - using axis, zones, tool, and long name
computation1 = lf.add_computation(
    "CMPT1", axis=[ax1], source=tool2, zones=(zone1, zone2, zone3), dimension=[2]
)
computation1.values.value = [[1, 2], [1, 3], [1, 4]]

computation2 = lf.add_computation("CMPT2", values=[2.3, 11.12312, 2231213.22])
computation3 = lf.add_computation("CMPT3", values=[3.14], long_name=long_name3)


# process - using channels, computations, and parameters
process1 = lf.add_process(
    "PROC",
    input_channels=(ch1, ch2),
    output_channels=(ch4,),
    input_computations=(computation1,),
    output_computations=(computation2, computation3),
    properties=[enums.Property.AVERAGED, enums.Property.LOCALLY_DEFINED],
)


# calibration coefficient
coef1 = lf.add_calibration_coefficient(
    "CC1",
    label="Gain",
    coefficients=[100.1, 100.2],
    references=[122, 123],
    plus_tolerances=[2, 2.5],
    minus_tolerances=[3, 2.4],
)
coef2 = lf.add_calibration_coefficient("CC2", coefficients=[1.2, 2.2, 3.4])
coef3 = lf.add_calibration_coefficient("CC3", coefficients=[2])


# calibration measurement - use Axis and Channel
cmeas1 = lf.add_calibration_measurement(
    "CM1",
    phase=enums.CalibrationMeasurementPhase.BEFORE,
    measurement_type="Plus",
    axis=ax1,
    measurement_source=ch1,
    measurement=[2.1, 2.5, 2.4],
    sample_count=3,
    begin_time=15,
    standard=[20, 25],
)
cmeas2 = lf.add_calibration_measurement(
    "CM2",
    measurement_source=ch5,
    measurement=[30, 40, 55, 61],
    begin_time=datetime.now() - timedelta(hours=20),
)


# calibration - using calibration coefficient & measurement, channels, and parameters
calibration1 = lf.add_calibration(
    "CALIB-1",
    calibrated_channels=(ch4,),
    uncalibrated_channels=(ch1, ch5),
    coefficients=(coef1, coef2, coef3),
    measurements=(cmeas1, cmeas2),
    parameters=(parameter2,),
    method="Two-point linear",
)

# well reference point
wrp = lf.add_well_reference_point(
    "WELL-REF",
    coordinate_1_name="Latitude",
    coordinate_1_value=40.34,
    coordinate_2_name="Longitude",
    coordinate_2_value=23.4232,
)


# path - using frame, well reference point, and channels
# Note: DeepView doesn't open files with Path defined, but dlisio opens them fine
# path1 = df.add_path('PATH-1', frame_type=main_frame, well_reference_point=wrp, value=(ch1, ch3, ch4),
#                     borehole_depth=122.12, vertical_depth=211.1, radial_drift=12, angular_drift=1.11, time=13)
# path2 = df.add_path('PATH-2', frame_type=second_frame, value=(ch5, ch6), tool_zero_offset=1231.1, time=11.1)


# message
message1 = lf.add_message(
    "MSG1", text=["Some message", "part 2 of the message"], time=datetime.now()
)
message2 = lf.add_message("MSG2", text=["You have a message"], message_type="Command")
message3 = lf.add_message(
    "MSG3", vertical_depth=213.1, text=["More", "text"], time=121.22
)


# comment
comment1 = lf.add_comment(
    "CMT1", text=["Part 1 of the comment", "Part 2 of the comment"]
)
comment2 = lf.add_comment("COMMENT-2", text=["Short comment"])


# no-format & no-format frame data
no_format1 = lf.add_no_format(
    "NF1", consumer_name="Some consumer", description="Some description"
)
no_format2 = lf.add_no_format("NF2", description="XYZ")
ndf1_1 = lf.add_no_format_frame_data(no_format1, data="XYZ")
nfd1_2 = lf.add_no_format_frame_data(no_format1, data="ABC")
nfd2_1 = lf.add_no_format_frame_data(no_format2, data="Lorem ipsum")


# groups
group1 = lf.add_group(
    "DEPTH ZONES", description="Zones describing depth", object_list=(zone1, zone3)
)
group2 = lf.add_group("MESSAGES", object_list=(message1, message2, message3))
group3 = lf.add_group("INDEX CHANNELS", object_list=(ch1, ch5))
lf.add_group("ALL CHANNELS", object_list=(ch2, ch3, ch4, ch6), group_list=(group3,))
group5 = lf.add_group("GROUPS", group_list=(group1, group2, group3))


# write the file
df.write("./tmp.DLIS", input_chunk_size=20, output_chunk_size=2**13)



from datetime import datetime, timedelta
import numpy as np
import pytest

from dliswriter.logical_record.eflr_types.origin import OriginItem, OriginSet
from dliswriter.utils.internal.struct_writer import ULONG_OFFSET
from dliswriter import high_compatibility_mode_decorator


def test_origin_creation() -> None:
    """Test creating OriginItem."""

    origin = OriginItem(
        "DEFAULT ORIGIN",
        parent=OriginSet(),
        origin_reference=1,
        file_set_number=1,
        creation_time="2050/03/02 15:30:00",
        file_id="WELL ID",
        file_set_name="Test file set name",
        file_number=8,
        run_number=13,
        well_id=5,
        well_name="Test well name",
        field_name="Test field name",
        company="Test company",
    )

    assert origin.name == "DEFAULT ORIGIN"

    assert origin.creation_time.value == datetime(
        year=2050, month=3, day=2, hour=15, minute=30
    )
    assert origin.file_id.value == "WELL ID"
    assert origin.file_set_name.value == "Test file set name"
    assert origin.file_set_number.value == 1
    assert origin.origin_reference == 1
    assert origin.run_number.value == 13
    assert origin.well_id.value == 5
    assert origin.well_name.value == "Test well name"
    assert origin.field_name.value == "Test field name"
    assert origin.company.value == "Test company"

    # not set - absent from config
    assert origin.producer_name.value is None
    assert origin.product.value is None
    assert origin.order_number.value is None
    assert origin.version.value is None
    assert origin.programs.value is None

    # OriginSet
    assert isinstance(origin.parent, OriginSet)
    assert origin.parent.set_name is None


def test_origin_creation_no_dtime_in_attributes() -> None:
    """Test that if creation_time is missing, the origin gets the current date and time as creation time."""

    origin = OriginItem(
        "Some origin name",
        well_name="Some well name",
        origin_reference=11,
        parent=OriginSet(),
    )

    assert origin.name == "Some origin name"
    assert origin.well_name.value == "Some well name"

    assert (
        timedelta(seconds=0)
        <= datetime.now() - origin.creation_time.value
        < timedelta(seconds=1)
    )


def test_origin_creation_no_attributes() -> None:
    """Test creating OriginItem with minimum number of parameters."""

    origin = OriginItem("Some origin name", origin_reference=1, parent=OriginSet())

    assert origin.name == "Some origin name"
    assert origin.well_name.value is None

    assert origin.file_set_number.value is not None
    assert isinstance(origin.file_set_number.value, int)
    assert 1 <= origin.file_set_number.value <= np.iinfo(np.uint32).max - ULONG_OFFSET

    assert (
        timedelta(seconds=0)
        <= datetime.now() - origin.creation_time.value
        < timedelta(seconds=1)
    )

    origin.make_item_body_bytes()  # this is where defaults are set
    assert origin.field_name.value == "WILDCAT"  # default according to RP66


def test_no_reassign_file_set_number() -> None:
    origin = OriginItem("Some origin name", origin_reference=1, parent=OriginSet())

    with pytest.raises(
        RuntimeError, match="File set number should not be reassigned.*"
    ):
        origin.file_set_number.value = 15


@high_compatibility_mode_decorator
def test_file_set_number_high_compat_mode() -> None:
    parent = OriginSet()

    origin1 = OriginItem("ORIGIN", origin_reference=1, parent=parent)
    assert origin1.file_set_number.value == 1

    origin2 = OriginItem(
        "ANOTHER-ORIGIN", origin_reference=2, parent=parent, file_set_number=15
    )
    assert origin2.file_set_number.value == 15

    origin3 = OriginItem("ORIG", origin_reference=3, parent=parent)
    assert origin3.file_set_number.value == 3


@pytest.mark.parametrize("name", ("MY-ORIGIN", "ORI_124", "421ORIGIN5"))
@high_compatibility_mode_decorator
def test_name_compatible(name: str) -> None:
    OriginItem(
        name, origin_reference=1, parent=OriginSet()
    )  # no error = name accepted, test passed


@pytest.mark.parametrize("name", ("Origin", "MY ORIGIN", "ORIGIN.3"))
@high_compatibility_mode_decorator
def test_name_not_compatible(name: str) -> None:
    with pytest.raises(
        ValueError,
        match=".*strings can contain only uppercase characters, digits, dashes, .*",
    ):
        OriginItem(name, origin_reference=1, parent=OriginSet())



from enum import Enum
from typing import Union, Optional, Callable
import logging

from dliswriter.configuration import global_config


logger = logging.getLogger(__name__)


class ValidatorEnum(str, Enum):
    """Define an enum with a converter defining ability.

    The 'make_converter' method returns a callable which verifies whether a provided value can be found among the
    enum's members or their values. It then returns the value of the relevant member or raises an error if no fitting
    member is found.

    The enum's values are expected to be strings.
    """

    value: str

    @classmethod
    def make_converter(cls, label: Optional[str] = None, allow_none: bool = False, soft: bool = False) -> Callable:
        def converter(v: Union[str, None, "ValidatorEnum"]) -> Union[str, None]:
            if allow_none and v is None:
                return None

            try:
                cls(v)
            except ValueError:
                pass
            else:
                if isinstance(v, cls):
                    return v.value
                return v

            if not isinstance(v, str):
                raise TypeError(f"Expected a str, got {type(v)}: {v}")

            message = (f"{repr(v)} is not one of the allowed {label or 'values'}: "
                       f"{', '.join(v.value for v in cls.__members__.values())}")
            if soft and not global_config.high_compat_mode:
                logger.warning(message)
                return v
            else:
                raise ValueError(message)

        return converter



import logging
from typing import Any

from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem, DimensionedItem
from dliswriter.utils.internal.internal_enums import EFLRType
from dliswriter.logical_record.eflr_types.zone import ZoneSet
from dliswriter.logical_record.eflr_types.axis import AxisSet
from dliswriter.logical_record.eflr_types.long_name import LongNameSet
from dliswriter.logical_record.core.attribute import Attribute, EFLRAttribute, DimensionAttribute, EFLROrTextAttribute
from dliswriter.utils.internal.value_checkers import convert_maybe_numeric


logger = logging.getLogger(__name__)


class ParameterItem(EFLRItem, DimensionedItem):
    """Model an object being part of Parameter EFLR."""

    parent: "ParameterSet"

    def __init__(self, name: str, parent: "ParameterSet", **kwargs: Any) -> None:
        """Initialise ParameterItem.

        Args:
            name        :   Name of the ParameterItem.
            parent      :   Parent ParameterSet of this ParameterItem.
            **kwargs    :   Values of to be set as characteristics of the ParameterItem Attributes.
        """

        self.long_name = EFLROrTextAttribute('long_name', object_class=LongNameSet)
        self.dimension = DimensionAttribute('dimension')
        self.axis = EFLRAttribute('axis', object_class=AxisSet, multivalued=True)
        self.zones = EFLRAttribute('zones', object_class=ZoneSet, multivalued=True)
        self.values = Attribute('values', converter=convert_maybe_numeric, multivalued=True,
                                multidimensional=True)

        super().__init__(name, parent=parent, **kwargs)

    def _run_checks_and_set_defaults(self) -> None:
        """Set default values of some attributes if no values have been set so far."""

        if self.values.value is not None:
            if self.zones.value is not None:
                if (nv := len(self.values.value)) != (nz := self.zones.count):
                    raise RuntimeError("A Parameter must have the same number of values and zones if both are "
                                       f"defined; got {nv} channels and {nz} zones in {self}")
            else:
                cv = self.values.count
                if cv is not None and cv > 1:
                    raise ValueError(f"{self} does not have any zones defined, so only a single value is permitted; "
                                     f"got {cv}: {self.values.value}")

        self._check_axis_vs_dimension()
        self._check_or_set_value_dimensionality(self.values.value)

        if self.values.value and not self.dimension.value:
            logger.debug(f"Setting dimension of '{self}' to [1]")
            self.dimension.value = [1]


class ParameterSet(EFLRSet):
    """Model Parameter EFLR."""

    set_type = 'PARAMETER'
    logical_record_type = EFLRType.STATIC
    item_type = ParameterItem


ParameterItem.parent_eflr_class = ParameterSet



import os
from typing import Union
import numpy as np

from dliswriter.file import DLISFile
from dliswriter.logical_record import eflr_types
from dliswriter.utils.enums import Unit

from tests.dlis_files_for_testing.common import make_df


def _add_channels(
    df: DLISFile, lf: int
) -> tuple[eflr_types.ChannelItem, eflr_types.ChannelItem]:
    ch_depth = df.logical_files[lf].add_channel(
        name="depth",
        dataset_name="/contents/depth",
        units=Unit.METER,
        cast_dtype=np.float64,
    )

    ch_rpm = df.logical_files[lf].add_channel(
        name="surface rpm",
        dataset_name="contents/rpm",
        cast_dtype=np.float64,
        dimension=[1],
    )

    return ch_depth, ch_rpm


def _add_frame(
    df: DLISFile, lf: int, channels: tuple[eflr_types.ChannelItem, ...]
) -> eflr_types.FrameItem:
    fr = df.logical_files[lf].add_frame(
        name="MAIN", index_type="DEPTH", channels=channels
    )

    fr.spacing.units = "m"

    return fr


def create_dlis_file_object() -> DLISFile:
    df = make_df()

    logical_file_index = 0
    channels = _add_channels(df, logical_file_index)
    _add_frame(df, logical_file_index, channels)

    return df


def write_depth_based_dlis(
    fname: Union[str, os.PathLike[str]], data: Union[dict, os.PathLike[str], np.ndarray]
) -> None:
    df = create_dlis_file_object()
    df.write(fname, data=data)



from abc import abstractmethod
import logging
from typing import Union, Any

from dliswriter.logical_record.core.logical_record.logical_record_bytes import LogicalRecordBytes
from dliswriter.utils.internal.internal_enums import RepresentationCode, EFLRType, IFLRType


logger = logging.getLogger(__name__)


class LRMeta(type):
    """Define a metaclass for all logical records.

    This metaclass adds a class-level property 'lr_type_struct' to LogicalRecord. This allows the value to be computed
    once for each LogicalRecord subclass.
    """

    logical_record_type: Union[EFLRType, IFLRType]
    _lr_type_struct: bytes

    def __new__(cls, *args: Any, **kwargs: Any) -> "LRMeta":
        """Create a new LogicalRecord (sub)class.

        Add a class-level attribute _lr_type_struct to be set when the value is first accessed through the corresponding
        property.
        """

        obj: LRMeta = super().__new__(cls, *args, **kwargs)
        obj._lr_type_struct = b''
        return obj

    @property
    def lr_type_struct(cls) -> bytes:
        """Bytes describing the type of the logical record."""

        if not cls._lr_type_struct:
            cls._lr_type_struct = RepresentationCode.USHORT.convert(cls.logical_record_type.value)
        return cls._lr_type_struct


class LogicalRecord(metaclass=LRMeta):
    """Create a base for all logical record classes."""

    is_eflr: bool = NotImplemented  #: indication whether this is an explicitly or indirectly formatted LR
    logical_record_type: Union[EFLRType, IFLRType] = NotImplemented  #: int-enum denoting type of the logical record

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialise a LogicalRecord."""

        pass

    @abstractmethod
    def _make_body_bytes(self) -> bytes:
        """Create bytes describing the body of this LogicalRecord.

        See the implementations subclasses: EFLR, FrameData (IFLR), and NoFormatFrameData (IFLR).
        """

        pass

    def represent_as_bytes(self) -> LogicalRecordBytes:
        """Create bytes representing the LogicalRecord.

        The bytes are wrapped in LogicalRecordBytes object which facilitates dividing the array of bytes into segments
        which can fit into visible records (physical units of DLIS).
        """

        return LogicalRecordBytes(
            self._make_body_bytes(),
            lr_type_struct=self.__class__.lr_type_struct,
            is_eflr=self.is_eflr
        )



from datetime import datetime, timezone
from dlisio import dlis    # type: ignore  # untyped library

from tests.common import check_list_of_objects


def test_calibration_measurement_params(short_dlis: dlis.file.LogicalFile) -> None:
    """Check attributes of CalibrationMeasurement object in the DLIS file."""

    m = short_dlis.measurements[0]

    assert m.name == "CMEASURE-1"
    assert m.phase == 'BEFORE'
    assert m.axis[0].name == 'Axis-4'
    assert m.source.name == "Channel 1"
    assert m.mtype == 'Plus'
    assert m.samples.tolist() == [[12.2323, 12.2131]]
    assert m.samplecount == 12
    assert m.max_deviation.tolist() == [2.2324, 3.121]
    assert m.begin_time.replace(tzinfo=timezone.utc) == datetime(2050, 3, 12, 12, 30).astimezone(timezone.utc)
    assert m.duration == 15
    assert m.reference.tolist() == [11, 12]
    assert m.standard == [11.2, 12.2]
    assert m.plus_tolerance.tolist() == [2, 2]
    assert m.minus_tolerance.tolist() == [1, 2]
    assert m.origin == 42
    assert m.dimension == [2]


def test_calibration_coefficient_params(short_dlis: dlis.file.LogicalFile) -> None:
    """Check attributes of CalibrationCoefficient object in the DLIS file."""

    c = short_dlis.coefficients[0]

    assert c.name == "COEF-1"
    assert c.label == 'Gain'
    assert c.coefficients == [100.2, 201.3]
    assert c.references == [89, 298]
    assert c.plus_tolerance == [100.2, 222.124]
    assert c.minus_tolerance == [87.23, 214]
    assert c.origin == 42


def test_calibration_params(short_dlis: dlis.file.LogicalFile) -> None:
    """Check attributes of Calibration object in the DLIS file."""

    c = short_dlis.calibrations[0]

    assert c.name == 'CALIB-MAIN'
    assert c.origin == 42

    check_list_of_objects(c.calibrated, ("Channel 1", "Channel 2"))
    check_list_of_objects(c.uncalibrated, ("amplitude", "radius", "radius_pooh"))
    check_list_of_objects(c.coefficients, ("COEF-1",))
    check_list_of_objects(c.measurements, ("CMEASURE-1",))
    check_list_of_objects(c.parameters, ("Param-1", "Param-2", "Param-3"))



import numpy as np
from pathlib import Path
import pytest
import logging
from typing import Union, Any

from dliswriter.utils.source_data_wrappers import DictDataWrapper, SourceDataWrapper


source_data_type = dict[str, np.ndarray]


@pytest.fixture
def data() -> source_data_type:
    """Mock source data for dict data wrapper."""

    n = 100
    d = {
        'depth': np.arange(n) * 0.1,
        'rpm': (10 * np.random.rand(n)).astype(np.int32),
        'amplitude': np.random.rand(n, 128).astype(np.float32),
    }

    return d


def test_basic_properties(data: source_data_type) -> None:
    """Check basic attributes of a dict data wrapper instance created with default settings."""

    w = DictDataWrapper(data)
    assert w.data_source is data
    assert w.n_rows == 100
    assert isinstance(w.dtype, np.dtype)
    assert isinstance(w.dtype.names, tuple)
    assert len(w.dtype.names) == 3


def test_creation_without_mapping(data: source_data_type) -> None:
    """Check that a dict wrapper with default mapping is created correctly."""

    w = DictDataWrapper(data)

    assert w.dtype.names == ('depth', 'rpm', 'amplitude')
    assert w.dtype[0] == np.float64
    assert w.dtype[1] == np.int32
    assert w.dtype[2] == (np.float32, (128,))


def test_creation_with_mapping(data: source_data_type) -> None:
    """Check that a dict wrapper with alternative mapping is created correctly."""

    w = DictDataWrapper(data, mapping={'MD': 'depth', 'RPM': 'rpm', 'AMP': 'amplitude'})

    assert w.dtype.names == ('MD', 'RPM', 'AMP')
    assert w.dtype[0] == np.float64
    assert w.dtype[1] == np.int32
    assert w.dtype[2] == (np.float32, (128,))


def test_creation_with_mapping_omitting_dsets(data: source_data_type) -> None:
    """Test creating a dict wrapper which omits some of the datasets found in the source data."""

    w = DictDataWrapper(data, mapping={'MD': 'depth', 'AMP': 'amplitude'})  # no rpm

    assert w.dtype.names == ('MD', 'AMP')
    assert w.dtype[0] == np.float64
    assert w.dtype[1] == (np.float32, (128,))


def test_creation_with_known_dtypes(data: source_data_type) -> None:
    """Test creating a dict wrapper with some data types explicitly provided."""

    w = DictDataWrapper(data, known_dtypes={'depth': np.float32, 'rpm': np.float64})

    assert w.dtype[0] == np.float32
    assert w.dtype[1] == np.float64
    assert w.dtype[2] == (np.float32, (128,))


def test_creation_with_known_dtypes_and_mapping(data: source_data_type) -> None:
    """Test creating a dict wrapper with alternative mapping and provided (matching) data types."""

    w = DictDataWrapper(
        data,
        mapping={'MD': 'depth', 'RPM': 'rpm', 'AMP': 'amplitude'},
        known_dtypes={'MD': np.float32, 'AMP': np.float64}  # names the same as keys of the mapping dict
    )

    assert w.dtype[0] == np.float32
    assert w.dtype[1] == np.int32
    assert w.dtype[2] == (np.float64, (128,))


@pytest.mark.parametrize('data_dict', (np.arange(12), [10, 20, 30, 40], 'path/to/something'))
def test_type_error_if_not_dict(data_dict: Any) -> None:
    """Check that a type error is raised if source data are not a dict."""

    with pytest.raises(TypeError, match="Expected a dictionary.*"):
        DictDataWrapper(data_dict)


@pytest.mark.parametrize('radius', (3, object(), 'path/to/something', (1, 2, 4.5), list(range(100))))
def test_type_error_if_not_numpy(data: source_data_type, radius: Any) -> None:
    """Check that a type error is raised if values of the data dict are anything other than a numpy.ndarray."""

    data2 = data | {'radius': radius}

    with pytest.raises(TypeError, match='Dict values must be numpy arrays.*'):
        DictDataWrapper(data2)


@pytest.mark.parametrize('key', (3, Path(__file__).resolve(), (1, 2, 3)))
def test_type_error_if_key_not_str(data: source_data_type, key: Any) -> None:
    """Check that a type error is raised if keys of the data dictionary are not strings."""

    data2 = data | {key: np.random.rand(100)}

    with pytest.raises(TypeError, match="Source dictionary keys must be strings.*"):
        DictDataWrapper(data2)


def test_creation_from_superclass(data: source_data_type) -> None:
    """Check that a dict wrapper is created from SourceDataWrapper.make_wrapper if the source data are a dict."""

    w = SourceDataWrapper.make_wrapper(data)
    assert isinstance(w, DictDataWrapper)


@pytest.mark.parametrize(('from_idx', 'to_idx', 'n_rows'), ((0, 10, 10), (60, 63, 3), (92, None, 8)))
def test_creation_with_from_and_to_idx(data: source_data_type, from_idx: int, to_idx: Union[int, None],
                                       n_rows: int) -> None:
    """Check creating a dict wrapper with specified from- and to-indices in the original data."""

    w = DictDataWrapper(data, from_idx=from_idx, to_idx=to_idx)
    assert w.n_rows == n_rows


@pytest.mark.parametrize(('start', 'stop'), ((0, 10), (1, 2), (3, 3), (3, 17), (61, 100)))
def test_load_chunk(data: source_data_type, start: int, stop: int) -> None:
    """Test loading a data chunk from a dict wrapper."""

    w = DictDataWrapper(data)
    chunk = w.load_chunk(start, stop)

    assert chunk.size == stop - start
    assert chunk.dtype == w.dtype
    assert (chunk['depth'] == data['depth'][start:stop]).all()
    assert (chunk['rpm'] == data['rpm'][start:stop]).all()
    assert (chunk['amplitude'] == data['amplitude'][start:stop]).all()


@pytest.mark.parametrize('start', (0, 10, 99, 100))
def test_load_chunk_until_end(data: source_data_type, start: int) -> None:
    """Test loading a data chunk from a specified point until the end."""

    w = DictDataWrapper(data)
    chunk = w.load_chunk(start, stop=None)

    assert chunk.size == 100 - start
    assert chunk.dtype == w.dtype
    assert (chunk['depth'] == data['depth'][start:]).all()
    assert (chunk['rpm'] == data['rpm'][start:]).all()
    assert (chunk['amplitude'] == data['amplitude'][start:]).all()


@pytest.mark.parametrize(('start', 'stop'), ((0, 101), (99, 1000)))
def test_load_chunk_stop_too_large(data: source_data_type, start: int, stop: int) -> None:
    """Check that a ValueError is raised if the specified chunk stop index is too large."""

    w = DictDataWrapper(data)

    with pytest.raises(ValueError, match=f"Cannot load chunk up to row {stop}.*"):
        w.load_chunk(start, stop)


@pytest.mark.parametrize('start', (101, 1210))
def test_load_chunk_start_too_large(data: source_data_type, start: int) -> None:
    """Check that a ValueError is raised if the specified chunk start index is too large."""

    w = DictDataWrapper(data)

    with pytest.raises(ValueError, match=f"Cannot load chunk from row {start}.*"):
        w.load_chunk(start, None)


@pytest.mark.parametrize(('start', 'stop'), ((10, 9), (90, 10)))
def test_load_chunk_start_larger_than_stop(data: source_data_type, start: int, stop: int) -> None:
    """Check that a ValueError is raised if the specified chunk start is larger than the stop."""

    w = DictDataWrapper(data)

    with pytest.raises(ValueError, match="Stop row cannot be smaller than start row.*"):
        w.load_chunk(start, stop)


@pytest.mark.parametrize(('start', 'from_idx'), ((60, 50), (10, 92), (91, 10)))
def test_load_chunk_start_too_large_with_from_idx(data: source_data_type, start: int, from_idx: int) -> None:
    """Check that a ValueError is raised if the specified chunk start index is too large with a from_idx specified."""

    w = DictDataWrapper(data, from_idx=from_idx)

    with pytest.raises(ValueError, match=f"Cannot load chunk from row {start}.*"):
        w.load_chunk(start, None)


@pytest.mark.parametrize(('stop', 'from_idx'), ((60, 50), (10, 92), (91, 10)))
def test_load_chunk_stop_too_large_with_from_idx(data: source_data_type, stop: int, from_idx: int) -> None:
    """Check that a ValueError is raised if the specified chunk stop index is too large with a from_idx specified."""

    w = DictDataWrapper(data, from_idx=from_idx)

    with pytest.raises(ValueError, match=f"Cannot load chunk up to row {stop}.*"):
        w.load_chunk(0, stop)


@pytest.mark.parametrize(('start', 'stop', 'from_idx', 'to_idx'), (
        (0, 10, 20, 40),
        (1, 2, 90, 92),
        (3, 3, 28, 55),
        (3, 17, 83, None)
))
def test_load_chunk_with_from_and_to_idx(data: source_data_type, start: int, stop: int, from_idx: int,
                                         to_idx: Union[int, None]) -> None:
    """Test loading a data chunk if from_idx and to_idx were specified."""

    w = DictDataWrapper(data, from_idx=from_idx, to_idx=to_idx)
    chunk = w.load_chunk(start, stop)

    assert chunk.size == stop - start
    assert chunk.dtype == w.dtype
    assert (chunk['depth'] == data['depth'][from_idx+start:from_idx+stop]).all()
    assert (chunk['rpm'] == data['rpm'][from_idx+start:from_idx+stop]).all()
    assert (chunk['amplitude'] == data['amplitude'][from_idx+start:from_idx+stop]).all()


@pytest.mark.parametrize(('start', 'from_idx', 'to_idx'), (
        (0, 20, 40),
        (1, 90, 92),
        (3, 28, 55),
        (3, 83, None)
))
def test_load_chunk_until_end_with_from_and_to_idx(data: source_data_type, start: int, from_idx: int,
                                                   to_idx: Union[int, None]) -> None:
    """Test loading a data chunk until the end if from_idx and to_idx were specified."""

    w = DictDataWrapper(data, from_idx=from_idx, to_idx=to_idx)
    chunk = w.load_chunk(start, stop=None)

    assert chunk.size == ((to_idx or 100) - from_idx) - start
    assert chunk.dtype == w.dtype
    assert (chunk['depth'] == data['depth'][from_idx+start:to_idx]).all()
    assert (chunk['rpm'] == data['rpm'][from_idx+start:to_idx]).all()
    assert (chunk['amplitude'] == data['amplitude'][from_idx+start:to_idx]).all()


def test_load_chunk_alternative_mapping(data: source_data_type) -> None:
    """Test loading a data chunk if alternative data names mapping was provided."""

    w = DictDataWrapper(data, mapping={'AMP': 'amplitude', 'MD': 'depth'})  # no rpm, switched order
    chunk = w.load_chunk(10, 20)

    assert chunk.size == 10
    assert chunk.dtype.names == ('AMP', 'MD')
    assert chunk.dtype == w.dtype

    assert (chunk['AMP'] == data['amplitude'][10:20]).all()
    assert (chunk['MD'] == data['depth'][10:20]).all()


@pytest.mark.parametrize(("from_idx", "to_idx", "n_rows"), ((0, None, 100), (20, 60, 40)))
def test_chunked_generator_all_rows(data: source_data_type, caplog: pytest.LogCaptureFixture, from_idx: int,
                                    to_idx: Union[int, None], n_rows: int) -> None:
    """Test loading data in a single chunk."""

    w = DictDataWrapper(data, from_idx=from_idx, to_idx=to_idx)

    with caplog.at_level(logging.DEBUG, logger='dliswriter'):
        gen = w.make_chunked_generator(chunk_rows=None)
        rows = list(gen)

    assert f"Data will be loaded in a single chunk of {n_rows}" in caplog.text
    assert "Loading chunk 1/1" in caplog.text

    assert len(rows) == n_rows

    assert rows[0]['depth'] == data['depth'][from_idx:to_idx][0]
    assert (rows[17]['amplitude'] == data['amplitude'][from_idx:to_idx][17]).all()
    assert rows[-2]['rpm'] == data['rpm'][from_idx:to_idx][n_rows-2]


@pytest.mark.parametrize(("chunk_rows", "n_chunks"), ((10, 10), (25, 4)))
def test_chunked_generator_full_chunks(data: source_data_type, caplog: pytest.LogCaptureFixture, chunk_rows: int,
                                       n_chunks: int) -> None:
    """Test loading data in several full chunks."""

    w = DictDataWrapper(data)

    with caplog.at_level(logging.DEBUG, logger='dliswriter'):
        gen = w.make_chunked_generator(chunk_rows=chunk_rows)
        rows = list(gen)

    assert f"Data will be loaded in {n_chunks} chunk(s) of {chunk_rows} rows" in caplog.text
    assert "plus a last, smaller chunk" not in caplog.text

    for i in range(n_chunks):
        assert f"Loading chunk {i+1}/{n_chunks}" in caplog.text

    assert len(rows) == 100

    assert rows[15]['depth'] == data['depth'][15]
    assert (rows[0]['amplitude'] == data['amplitude'][0]).all()
    assert rows[3]['rpm'] == data['rpm'][3]


@pytest.mark.parametrize(("chunk_rows", "n_full_chunks", "remainder_rows"), ((15, 6, 10), (40, 2, 20), (60, 1, 40)))
def test_chunked_generator_remainder_chunk(data: source_data_type, caplog: pytest.LogCaptureFixture, chunk_rows: int,
                                           n_full_chunks: int, remainder_rows: int) -> None:
    """Test loading data in several full and last half-full chunk."""

    w = DictDataWrapper(data)

    with caplog.at_level(logging.DEBUG, logger='dliswriter'):
        gen = w.make_chunked_generator(chunk_rows=chunk_rows)
        rows = list(gen)

    assert f"Data will be loaded in {n_full_chunks} chunk(s) of {chunk_rows} rows" in caplog.text
    assert f"plus a last, smaller chunk of {remainder_rows} rows" in caplog.text

    for i in range(n_full_chunks + 1):
        assert f"Loading chunk {i+1}/{n_full_chunks + 1}" in caplog.text

    assert len(rows) == 100

    assert rows[80]['depth'] == data['depth'][80]
    assert (rows[2]['amplitude'] == data['amplitude'][2]).all()
    assert rows[24]['rpm'] == data['rpm'][24]


@pytest.mark.parametrize(("from_idx", "to_idx"), ((0, None), (10, 30)))
def test_getitem_default_mapping(data: source_data_type, from_idx: int, to_idx: Union[int, None]) -> None:
    """Test the __getitem__ functionality for a dict wrapper created with default mapping."""

    w = DictDataWrapper(data, from_idx=from_idx, to_idx=to_idx)

    for key in ('depth', 'amplitude', 'rpm'):
        assert isinstance(w[key], np.ndarray)

    assert (w['depth'] == data['depth'][from_idx:to_idx]).all()
    assert (w['amplitude'] == data['amplitude'][from_idx:to_idx]).all()
    assert (w['rpm'] == data['rpm'][from_idx:to_idx]).all()


@pytest.mark.parametrize(("from_idx", "to_idx"), ((0, None), (20, 21)))
def test_getitem_alternative_mapping(data: source_data_type, from_idx: int, to_idx: Union[int, None]) -> None:
    """Test the __getitem__ functionality for a dict wrapper created with alternative mapping."""

    w = DictDataWrapper(data, from_idx=from_idx, to_idx=to_idx, mapping={'MD': 'depth', 'RPM': 'rpm'})

    for key in ('MD', 'RPM'):
        assert isinstance(w[key], np.ndarray)

    assert (w['MD'] == data['depth'][from_idx:to_idx]).all()
    assert (w['RPM'] == data['rpm'][from_idx:to_idx]).all()

    for key in ('depth', 'rpm', 'amplitude'):
        with pytest.raises(ValueError, match=f"No dataset '{key}' found in the source data"):
            w[key]



import numpy as np
from typing import TYPE_CHECKING, Optional

from dliswriter.logical_record.core.iflr import IFLR
from dliswriter.utils.internal.struct_writer import write_struct_uvari
from dliswriter.utils.internal.internal_enums import IFLRType

if TYPE_CHECKING:
    from dliswriter.logical_record.eflr_types.frame import FrameItem


class FrameData(IFLR):
    """Model FrameData - an Implicitly Formatted Logical Record meant for storing numerical data.

    Each FrameData record begins with a reference to the FrameObject that it belongs to.
    """

    logical_record_type = IFLRType.FDATA

    def __init__(self, frame: "FrameItem", frame_number: int, slots: np.ndarray,
                 origin_reference: Optional[int] = None):
        """Initialise a FrameData.

        Args:
            frame               :   The frame that the data belongs to.
            frame_number        :   Index of the frame, i.e. index (starting from 1) of the row of data.
            slots               :   Data belonging to the object. It should be a row of a structured numpy array
                                    with consecutive items corresponding to the channels in the frame.
            origin_reference    :   Origin reference for the object.
        """

        super().__init__()

        self._frame = frame
        self._frame_number = frame_number
        self._slots = slots

        self.origin_reference = origin_reference

    def _make_body_bytes(self) -> bytes:
        """Create bytes describing the body of the FrameData object.

        This includes: reference to the parent frame, placement of the FrameData object (the index), and the data.
        """

        body = self._frame.obname + write_struct_uvari(self._frame_number)

        for s in self._slots:
            body += s.byteswap().tobytes()

        return body
